from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraDeviceProject(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_devices_projects'
    id_device_project = Column(Integer, Sequence('id_device_project_sequence'), primary_key=True,
                               autoincrement=True)
    id_device = Column(Integer, ForeignKey("t_devices.id_device", ondelete='cascade'), nullable=False)
    id_project = Column(Integer, ForeignKey("t_projects.id_project", ondelete='cascade'), nullable=False)

    device_project_project = relationship("TeraProject", viewonly=True)
    device_project_device = relationship("TeraDevice", viewonly=True)

    def to_json(self, ignore_fields=[], minimal=False):
        ignore_fields.extend(['device_project_project', 'device_project_device'])

        if minimal:
            ignore_fields.extend([])

        rval = super().to_json(ignore_fields=ignore_fields)

        return rval

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraDevice import TeraDevice
            from opentera.db.models.TeraProject import TeraProject
            device1 = TeraDevice.get_device_by_name('Apple Watch #W05P1')
            device2 = TeraDevice.get_device_by_name('Kit Télé #1')
            device3 = TeraDevice.get_device_by_name('Robot A')
            project1 = TeraProject.get_project_by_projectname('Default Project #1')
            project2 = TeraProject.get_project_by_projectname('Default Project #2')
            project3 = TeraProject.get_project_by_projectname('Secret Project #1')

            dev_proj = TeraDeviceProject()
            dev_proj.id_device = device1.id_device
            dev_proj.id_project = project1.id_project
            TeraDeviceProject.db().session.add(dev_proj)

            dev_proj = TeraDeviceProject()
            dev_proj.id_device = device2.id_device
            dev_proj.id_project = project1.id_project
            TeraDeviceProject.db().session.add(dev_proj)

            dev_proj = TeraDeviceProject()
            dev_proj.id_device = device1.id_device
            dev_proj.id_project = project3.id_project
            TeraDeviceProject.db().session.add(dev_proj)

            TeraDeviceProject.db().session.commit()

    @staticmethod
    def get_device_project_by_id(device_project_id: int, with_deleted: bool = False):
        return TeraDeviceProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_device_project=device_project_id).first()

    @staticmethod
    def get_device_project_id_for_device_and_project(device_id: int, project_id: int, with_deleted: bool = False):
        return TeraDeviceProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_project=project_id, id_device=device_id).first()

    @staticmethod
    def get_devices_for_project(project_id: int, with_deleted: bool = False):
        return TeraDeviceProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_project=project_id).all()

    @staticmethod
    def get_projects_for_device(device_id: int, with_deleted: bool = False):
        return TeraDeviceProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_device=device_id).all()

    @staticmethod
    def delete_with_ids(device_id: int, project_id: int, autocommit: bool = True):
        delete_obj = TeraDeviceProject.query.filter_by(id_device=device_id, id_project=project_id).first()
        if delete_obj:
            TeraDeviceProject.delete(delete_obj.id_device_project, autocommit=autocommit)

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        from opentera.db.models.TeraDeviceParticipant import TeraDeviceParticipant
        from opentera.db.models.TeraParticipant import TeraParticipant
        from opentera.db.models.TeraSession import TeraSession

        if TeraDeviceParticipant.query.execution_options(include_deleted=with_deleted).join(TeraParticipant).\
            filter(TeraParticipant.id_project == self.id_project).\
                filter(TeraDeviceParticipant.id_device == self.id_device).count():
            return IntegrityError('Project still has participant associated to the device',
                                  self.id_device_project, 't_participants')

        # Find sessions with matching device and project
        device_sessions = TeraSession.get_sessions_for_device(self.id_device, with_deleted=with_deleted)
        device_project_sessions = [ses.id_session for ses in device_sessions
                                   if ses.get_associated_project_id() == self.id_project]

        if len(device_project_sessions) > 0:
            return IntegrityError('Device still has sessions in this project',
                                  self.id_device_project, 't_sessions')

        return None

    @classmethod
    def insert(cls, dp):
        # Check if that site of that project has the site associated to it
        from opentera.db.models.TeraDeviceSite import TeraDeviceSite
        from opentera.db.models.TeraProject import TeraProject
        project = TeraProject.get_project_by_id(project_id=dp.id_project)

        if not project:
            raise IntegrityError(params='Project not found',
                                 orig='TeraDeviceProject.insert', statement='insert')

        device_site = TeraDeviceSite.get_device_site_id_for_device_and_site(site_id=project.id_site,
                                                                            device_id=dp.id_device)

        if not device_site:
            raise IntegrityError(params='Device not associated to project site',
                                 orig='TeraDeviceProject.insert', statement='insert')
        inserted_obj = super().insert(dp)
        return inserted_obj

    @classmethod
    def update(cls, update_id: int, values: dict):
        return



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Result.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cResult.proto\x12\x11opentera.protobuf\"\xb2\x01\n\x06Result\x12\x32\n\x04type\x18\x01 \x01(\x0e\x32$.opentera.protobuf.Result.ResultType\x12\x0c\n\x04\x63ode\x18\x02 \x01(\x03\x12\x0f\n\x07message\x18\x03 \x01(\t\"U\n\nResultType\x12\x12\n\x0eRESULT_UNKNOWN\x10\x00\x12\r\n\tRESULT_OK\x10\x01\x12\x10\n\x0cRESULT_ERROR\x10\x02\x12\x12\n\x0eRESULT_WARNING\x10\x03\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'Result_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _RESULT._serialized_start=36
  _RESULT._serialized_end=214
  _RESULT_RESULTTYPE._serialized_start=129
  _RESULT_RESULTTYPE._serialized_end=214
# @@protoc_insertion_point(module_scope)



from opentera.db.Base import BaseModel
from sqlalchemy import Column, Integer, String, Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraDeviceType(BaseModel):

    __tablename__ = 't_devices_types'
    id_device_type = Column(Integer, Sequence('id_device_type_sequence'), primary_key=True, autoincrement=True)
    device_type_name = Column(String, nullable=False)
    device_type_key = Column(String, nullable=False, unique=True)

    device_type_devices = relationship("TeraDevice", passive_deletes=True, back_populates='device_type')

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        device_type_json = super().to_json(ignore_fields=ignore_fields)

        return device_type_json

    def get_name(self) -> str:
        return self.device_type_name

    @staticmethod
    def create_defaults(test=False):
        if not test:
            return
        # For now Defaults are hard coded
        VIDEOCONFERENCE = TeraDeviceType()
        VIDEOCONFERENCE.device_type_name = 'Vidéoconférence'
        VIDEOCONFERENCE.device_type_key = 'videoconference'
        TeraDeviceType.db().session.add(VIDEOCONFERENCE)

        CAPTEUR = TeraDeviceType()
        CAPTEUR.device_type_name = 'Capteur'
        CAPTEUR.device_type_key = 'capteur'
        TeraDeviceType.db().session.add(CAPTEUR)

        ROBOT = TeraDeviceType()
        ROBOT.device_type_name = 'Robot'
        ROBOT.device_type_key = 'robot'
        TeraDeviceType.db().session.add(ROBOT)

        BUREAU_ACTIF = TeraDeviceType()
        BUREAU_ACTIF.device_type_name = 'Bureau Actif'
        BUREAU_ACTIF.device_type_key = 'bureau_actif'
        TeraDeviceType.db().session.add(BUREAU_ACTIF)

    @staticmethod
    def get_devices_types():
        return TeraDeviceType.query.all()

    @staticmethod
    def get_device_type(dev_type: int):
        return TeraDeviceType.query.filter_by(id_device_type=dev_type).first()

    @staticmethod
    def get_device_type_by_id(dev_id: int):
        return TeraDeviceType.query.filter_by(id_device_type=dev_id).first()

    @staticmethod
    def get_device_type_by_name(dev_name: str):
        return TeraDeviceType.query.filter_by(device_type_name=dev_name).all()

    @staticmethod
    def get_device_type_by_key(dev_key: str):
        return TeraDeviceType.query.filter_by(device_type_key=dev_key).first()

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        # More efficient that relationships
        from opentera.db.models.TeraDevice import TeraDevice  # Here to prevent circular import
        if TeraDevice.get_count(filters={'id_device_type': self.id_device_type}, with_deleted=with_deleted) > 0:
            return IntegrityError('Device Type still has associated devices', self.id_device_type, 't_devices')
        return None

    @staticmethod
    def commit():
        TeraDeviceType.db().session.commit()




import string


class TeraForm:
    def __init__(self, object_name):
        self.sections = []
        self.object_name = object_name

    def add_section(self, section):
        self.sections.append(section)

    def to_dict(self):
        sections = []
        order = 1
        for section in self.sections:
            sections.append(section.to_dict(order))
            order += 1
        object_dict = {'objecttype': self.object_name,
                       'sections': sections}
        return object_dict


class TeraFormSection:

    def __init__(self, section_id: string, section_label: string):
        self.id = section_id
        self.label = section_label
        self.items = []

    def add_item(self, item):
        self.items.append(item)

    def to_dict(self, order: int = -1):
        section = {"id": self.id,
                   "label": self.label}

        if order >= 0:
            section['_order'] = order
        items = []
        item_order = 1
        for item in self.items:
            items.append(item.to_dict(item_order))
            item_order += 1
        section["items"] = items
        return section


class TeraFormItemCondition:

    # condition_item = any item in the form
    # condition_operator = operator of the form "=", "<>", ">" ...
    # condition_condition = value to check. Special flags: "changed"
    # hook = api call to get new values for that item when condition is true
    def __init__(self, condition_item: str,  condition_operator: str, condition_condition, hook: str = None):
        self.condition = condition_condition
        self.item = condition_item
        self.operator = condition_operator
        self.hook = hook

    def to_dict(self):
        item_condition = {"item": self.item,
                          "op": self.operator,
                          "condition": self.condition}
        if self.hook:
            item_condition["hook"] = self.hook
        return item_condition


class TeraFormValue:

    def __init__(self, value_id: string, value: string):
        self.id = value_id
        self.value = value

    def to_dict(self):
        base_dict = {"id": self.id, "value": self.value}
        return base_dict


class TeraFormItem:

    def __init__(self, item_id: str, item_label: str, item_type: str, item_required: bool = False,
                 item_values: list | None = None, item_default: str | None | bool = None,
                 item_condition: TeraFormItemCondition | None = None, item_options=None):
        if item_options is None:
            item_options = {}
        self.id = item_id
        self.label = item_label
        self.type = item_type
        self.required = item_required
        self.values = item_values
        self.default = item_default
        self.condition = item_condition
        self.options = item_options

    def set_condition(self, condition: TeraFormItemCondition):
        self.condition = condition

    def to_dict(self, order: int = -1):
        item = {"id": self.id,
                "label": self.label,
                "type": self.type,
                }
        if self.required:
            item["required"] = True

        if self.values:
            values = []
            id_index = 0
            for value in self.values:
                if isinstance(value, TeraFormValue):
                    values.append(value.to_dict())
                elif isinstance(values, list):
                    values.append(TeraFormValue(str(id_index), value).to_dict())
                    id_index += 1

            item["values"] = values

        if self.default:
            item["default"] = self.default

        if self.condition:
            item["condition"] = self.condition.to_dict()

        if order >= 0:
            item['_order'] = order

        if len(self.options) > 0:
            for option in self.options:
                item[option] = self.options[option]

        return item



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.models.TeraSession import TeraSession
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError
from enum import Enum, unique
from flask_babel import gettext


class TeraSessionType(BaseModel, SoftDeleteMixin):
    @unique
    class SessionCategoryEnum(Enum):
        SERVICE = 1
        DATACOLLECT = 2
        FILETRANSFER = 3
        PROTOCOL = 4

        def describe(self):
            return self.name, self.value

    __tablename__ = 't_sessions_types'
    id_session_type = Column(Integer, Sequence('id_session_type_sequence'), primary_key=True, autoincrement=True)
    id_service = Column(Integer, ForeignKey('t_services.id_service', ondelete='cascade'), nullable=True)
    session_type_name = Column(String, nullable=False, unique=False)
    session_type_online = Column(Boolean, nullable=False)
    session_type_config = Column(String, nullable=True)
    session_type_color = Column(String(7), nullable=False)
    session_type_category = Column(Integer, nullable=False)

    session_type_session_type_projects = relationship("TeraSessionTypeProject", viewonly=True)
    session_type_session_type_sites = relationship("TeraSessionTypeSite", viewonly=True)

    session_type_projects = relationship("TeraProject", secondary="t_sessions_types_projects",
                                         back_populates="project_session_types")
    session_type_sites = relationship("TeraSite", secondary="t_sessions_types_sites")

    session_type_service = relationship("TeraService")

    session_type_sessions = relationship("TeraSession", passive_deletes=True, back_populates='session_session_type')

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []
        ignore_fields.extend(['session_type_projects', 'session_type_devices_types', 'SessionCategoryEnum',
                              'session_type_service', 'session_type_sessions', 'session_type_session_type_projects',
                              'session_type_sites', 'session_type_session_type_sites'])
        if minimal:
            ignore_fields.extend(['session_type_online',
                                  'session_type_profile',
                                  'session_type_config'])
        rval = super().to_json(ignore_fields=ignore_fields)

        if not minimal:
            # Also includes service key and uuid
            if self.session_type_service:
                rval['session_type_service_key'] = self.session_type_service.service_key
                rval['session_type_service_uuid'] = self.session_type_service.service_uuid
        return rval

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_session_type': self.id_session_type, 'id_service': self.id_service}

    @staticmethod
    def create_defaults(test=False):
        if not test:
            return
        # from opentera.db.models.TeraProject import TeraProject
        # from opentera.db.models.TeraDeviceType import TeraDeviceType
        from opentera.db.models.TeraService import TeraService

        # type_project = TeraProject.get_project_by_projectname('Default Project #1')
        video_session = TeraSessionType()
        video_session.session_type_name = "Suivi vidéo"
        video_session.session_type_online = True
        video_session.session_type_config = ""
        video_session.session_type_color = "#00FF00"
        video_session.session_type_category = TeraSessionType.SessionCategoryEnum.SERVICE.value
        video_session.id_service = TeraService.get_service_by_key('VideoRehabService').id_service
        # video_session.session_type_projects = [type_project]
        # video_session.session_type_uses_devices_types = [TeraDeviceType.get_device_type(
        #     int(TeraDeviceType.DeviceTypeEnum.VIDEOCONFERENCE.value))]
        TeraSessionType.db().session.add(video_session)

        sensor_session = TeraSessionType()
        sensor_session.session_type_name = "Données Capteur"
        sensor_session.session_type_online = False
        sensor_session.session_type_config = ""
        sensor_session.session_type_color = "#0000FF"
        sensor_session.session_type_category = TeraSessionType.SessionCategoryEnum.FILETRANSFER.value
        # sensor_session.session_type_projects = [type_project]
        # sensor_session.session_type_uses_devices_types = [TeraDeviceType.get_device_type(
        #     int(TeraDeviceType.DeviceTypeEnum.SENSOR.value))]
        TeraSessionType.db().session.add(sensor_session)

        vsensor_session = TeraSessionType()
        vsensor_session.session_type_name = "Collecte données"
        vsensor_session.session_type_online = True
        vsensor_session.session_type_config = ""
        vsensor_session.session_type_color = "#00FFFF"
        # vsensor_session.session_type_projects = [type_project]
        vsensor_session.session_type_category = TeraSessionType.SessionCategoryEnum.DATACOLLECT.value
        # vsensor_session.session_type_uses_devices_types = [TeraDeviceType.get_device_type(
        #     int(TeraDeviceType.DeviceTypeEnum.SENSOR.value)), TeraDeviceType.get_device_type(
        #     int(TeraDeviceType.DeviceTypeEnum.VIDEOCONFERENCE.value))]
        TeraSessionType.db().session.add(vsensor_session)

        robot_session = TeraSessionType()
        robot_session.session_type_name = "Exercices individuels"
        robot_session.session_type_online = False
        robot_session.session_type_config = ""
        robot_session.session_type_color = "#FF00FF"
        # robot_session.session_type_projects = [type_project]
        robot_session.session_type_category = TeraSessionType.SessionCategoryEnum.PROTOCOL.value
        # robot_session.session_type_uses_devices_types = [TeraDeviceType.get_device_type(
        #     int(TeraDeviceType.DeviceTypeEnum.ROBOT.value))]
        TeraSessionType.db().session.add(robot_session)

        bureau_session = TeraSessionType()
        bureau_session.session_type_name = "FileTransfer"
        bureau_session.session_type_online = False
        bureau_session.session_type_config = ""
        bureau_session.session_type_color = "#FF00FF"
        bureau_session.session_type_category = TeraSessionType.SessionCategoryEnum.SERVICE.value
        # robot_session.session_type_uses_devices_types = [TeraDeviceType.get_device_type(
        #     int(TeraDeviceType.DeviceTypeEnum.ROBOT.value))]
        bureau_session.id_service = TeraService.get_service_by_key('FileTransferService').id_service
        TeraSessionType.db().session.add(bureau_session)

        TeraSessionType.db().session.commit()

    @staticmethod
    def get_session_type_by_id(ses_type_id: int, with_deleted: bool = False):
        return TeraSessionType.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_session_type=ses_type_id).first()

    @staticmethod
    def get_session_types_for_service(id_service: int, with_deleted: bool = False):
        return TeraSessionType.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service=id_service).all()

    @staticmethod
    def get_category_name(category: SessionCategoryEnum):
        name = gettext('Unknown')
        if category == TeraSessionType.SessionCategoryEnum.SERVICE:
            name = gettext('Service')
        if category == TeraSessionType.SessionCategoryEnum.FILETRANSFER:
            name = gettext('File Transfer')
        if category == TeraSessionType.SessionCategoryEnum.DATACOLLECT:
            name = gettext('Data Collect')
        if category == TeraSessionType.SessionCategoryEnum.PROTOCOL:
            name = gettext('Protocol')

        return name

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        if TeraSession.get_count(filters={'id_session_type': self.id_session_type}, with_deleted=with_deleted) > 0:
            return IntegrityError('Still have sessions with that type', self.id_session_type, 't_sessions')
        return None

    @classmethod
    def update(cls, id_st: int, values: dict):
        # Set service id to None if setted to 0
        if 'id_service' in values:
            if values['id_service'] == 0:
                del values['id_service']

        super().update(id_st, values)

    @classmethod
    def insert(cls, st):
        if st.id_service == 0:
            st.id_service = None

        super().insert(st)

    def hard_delete_before(self):
        # Delete sessions related to that session type
        for ses in self.session_type_sessions:
            ses.hard_delete()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: JoinSessionReplyEvent.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bJoinSessionReplyEvent.proto\x12\x11opentera.protobuf\"\xb3\x02\n\x15JoinSessionReplyEvent\x12\x14\n\x0csession_uuid\x18\x01 \x01(\t\x12\x11\n\tuser_uuid\x18\x02 \x01(\t\x12\x18\n\x10participant_uuid\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65vice_uuid\x18\x04 \x01(\t\x12\x46\n\njoin_reply\x18\x05 \x01(\x0e\x32\x32.opentera.protobuf.JoinSessionReplyEvent.ReplyType\x12\x11\n\treply_msg\x18\x06 \x01(\t\"g\n\tReplyType\x12\x11\n\rREPLY_UNKNOWN\x10\x00\x12\x12\n\x0eREPLY_ACCEPTED\x10\x01\x12\x10\n\x0cREPLY_DENIED\x10\x02\x12\x0e\n\nREPLY_BUSY\x10\x03\x12\x11\n\rREPLY_TIMEOUT\x10\x04\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'JoinSessionReplyEvent_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _JOINSESSIONREPLYEVENT._serialized_start=51
  _JOINSESSIONREPLYEVENT._serialized_end=358
  _JOINSESSIONREPLYEVENT_REPLYTYPE._serialized_start=255
  _JOINSESSIONREPLYEVENT_REPLYTYPE._serialized_end=358
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: RPCMessage.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10RPCMessage.proto\x12\x11opentera.protobuf\"\xb3\x01\n\x05Value\x12\x16\n\x0c\x64ouble_value\x18\x01 \x01(\x01H\x00\x12\x15\n\x0b\x66loat_value\x18\x02 \x01(\x02H\x00\x12\x13\n\tint_value\x18\x03 \x01(\x03H\x00\x12\x14\n\nuint_value\x18\x04 \x01(\x04H\x00\x12\x14\n\nbool_value\x18\x05 \x01(\x08H\x00\x12\x15\n\x0b\x62ytes_value\x18\x06 \x01(\x0cH\x00\x12\x16\n\x0cstring_value\x18\x07 \x01(\tH\x00\x42\x0b\n\targ_value\"u\n\nRPCMessage\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\x10\n\x08reply_to\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\x04\x12\x11\n\ttimestamp\x18\x04 \x01(\x01\x12&\n\x04\x61rgs\x18\x05 \x03(\x0b\x32\x18.opentera.protobuf.Valueb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'RPCMessage_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _VALUE._serialized_start=40
  _VALUE._serialized_end=219
  _RPCMESSAGE._serialized_start=221
  _RPCMESSAGE._serialized_end=338
# @@protoc_insertion_point(module_scope)



import uuid

from requests import Response
from flask import request
from opentera.redis.RedisRPCClient import RedisRPCClient


class TeraParticipantClient:

    def __init__(self, token_dict: dict, token: str, config_man):
        self.__participant_uuid = token_dict['participant_uuid']
        self.__id_participant = token_dict['id_participant']
        self.__participant_token = token

        backend_hostname = config_man.backend_config["hostname"]
        backend_port = str(config_man.backend_config["port"])

        self.__backend_url = 'https://' + backend_hostname + ':' + backend_port
        self.__config_man = config_man
        self.__rpc_client = RedisRPCClient(config_man.redis_config)

    @property
    def participant_uuid(self):
        return self.__participant_uuid

    @participant_uuid.setter
    def participant_uuid(self, u_uuid: uuid):
        self.__participant_uuid = u_uuid

    @property
    def id_participant(self):
        return self.__id_participant

    @id_participant.setter
    def id_participant(self, id_p: int):
        self.__id_participant = id_p

    @property
    def participant_token(self):
        return self.__participant_token

    @participant_token.setter
    def participant_token(self, token: str):
        self.__participant_token = token

    def get_participant_infos(self) -> dict:
        # Get information from service rpc function since participant has no rights from the participants api
        #participant_info = self.__rpc_client.call_service(self.__config_man.service_config['name'],
        #                                                  'participant_info', self.__participant_uuid)
        #return participant_info

        response = self.do_get_request_to_backend('/api/participant/participants')
        if response.status_code == 200:
            return response.json()
        return {}

    def do_get_request_to_backend(self, path: str) -> Response:
        from requests import get
        request_headers = {'Authorization': 'OpenTera ' + self.__participant_token}
        # TODO: remove verify=False and check certificate
        backend_response = get(url=self.__backend_url + path, headers=request_headers, verify=False)
        return backend_response

    def __repr__(self):
        return '<TeraParticipantClient - UUID: ' + self.__participant_uuid \
               + ', Token: ' + self.__participant_token + '>'



from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError
from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.models.TeraProject import TeraProject


class TeraParticipantGroup(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_participants_groups'
    id_participant_group = Column(Integer, Sequence('id_participantgroup_sequence'), primary_key=True,
                                  autoincrement=True)
    id_project = Column(Integer, ForeignKey('t_projects.id_project', ondelete='cascade'), nullable=False)
    participant_group_name = Column(String, nullable=False, unique=False)

    participant_group_project = relationship('TeraProject', back_populates='project_participants_groups')
    participant_group_participants = relationship("TeraParticipant", back_populates='participant_participant_group',
                                                  passive_deletes=True, cascade='delete')

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []
        ignore_fields.extend(['participant_group_project', 'participant_group_participants'])
        rval = super().to_json(ignore_fields=ignore_fields)

        if not minimal:
            rval['project_name'] = self.participant_group_project.project_name
        # rval['id_site'] = self.participant_group_project.id_site
        return rval

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_participant_group': self.id_participant_group}

    @staticmethod
    def get_participant_group_by_group_name(name: str, with_deleted: bool = False):
        return TeraParticipantGroup.query.execution_options(include_deleted=with_deleted)\
            .filter_by(participant_group_name=name).first()

    @staticmethod
    def get_participant_group_by_id(group_id: int, with_deleted: bool = False):
        return TeraParticipantGroup.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_participant_group=group_id).first()

    @staticmethod
    def get_participant_group_for_project(project_id: int, with_deleted: bool = False):
        return TeraParticipantGroup.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_project=project_id).all()

    @staticmethod
    def create_defaults(test=False):
        if test:
            base_pgroup = TeraParticipantGroup()
            base_pgroup.participant_group_name = 'Default Participant Group A'
            base_pgroup.id_project = TeraProject.get_project_by_projectname('Default Project #1').id_project
            TeraParticipantGroup.db().session.add(base_pgroup)

            base_pgroup2 = TeraParticipantGroup()
            base_pgroup2.participant_group_name = 'Default Participant Group B'
            base_pgroup2.id_project = TeraProject.get_project_by_projectname('Default Project #2').id_project
            TeraParticipantGroup.db().session.add(base_pgroup2)
            TeraParticipantGroup.db().session.commit()

    @classmethod
    def update(cls, update_id: int, values: dict):
        # If group project changed, also changed project from all participants in that group
        if 'id_project' in values:
            updated_group: TeraParticipantGroup = TeraParticipantGroup.get_participant_group_by_id(update_id)
            if updated_group:
                for participant in updated_group.participant_group_participants:
                    participant.id_project = values['id_project']
        super().update(update_id=update_id, values=values)

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        for participant in self.participant_group_participants:
            cannot_be_deleted_exception = participant.delete_check_integrity(with_deleted=with_deleted)
            if cannot_be_deleted_exception:
                return IntegrityError('Participant group still has participant(s)', self.id_participant_group,
                                      't_participants')
        return None



from opentera.forms.TeraForm import *
from flask_babel import gettext
from opentera.db.models.TeraSessionType import TeraSessionType


class TeraSessionTypeConfigForm:

    @staticmethod
    def get_session_type_config_form(session_type: TeraSessionType):
        # Handle session type configs for non-services session types
        form = TeraForm("session_type_config")
        return form.to_dict()





from opentera.db.Base import BaseModel
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship
import string
from string import digits, ascii_lowercase, ascii_uppercase
import random
import uuid


class TeraServerSettings(BaseModel):
    __tablename__ = 't_server_settings'
    id_server_settings = Column(Integer, Sequence('id_server_settings_sequence'), primary_key=True, autoincrement=True)
    server_settings_name = Column(String, nullable=False, unique=True)
    server_settings_value = Column(String, nullable=False, unique=False)

    # Constants
    ServerDeviceTokenKey = "TokenEncryptionKey"
    ServerParticipantTokenKey = "ParticipantTokenEncryptionKey"
    ServerUUID = "ServerUUID"
    ServerVersions = "ServerVersions"
    ServerDeviceRegisterKey = "DeviceRegisterKey"

    @staticmethod
    def create_defaults(test=False):
        # Create defaults settings
        # Token Encryption Key
        TeraServerSettings.set_server_setting(TeraServerSettings.ServerDeviceTokenKey,
                                              TeraServerSettings.generate_token_key(32))

        TeraServerSettings.set_server_setting(TeraServerSettings.ServerParticipantTokenKey,
                                              TeraServerSettings.generate_token_key(32))

        TeraServerSettings.set_server_setting(TeraServerSettings.ServerDeviceRegisterKey,
                                              TeraServerSettings.generate_token_key(10))

        # Unique server id
        server_uuid = str(uuid.uuid4())
        TeraServerSettings.set_server_setting(TeraServerSettings.ServerUUID, server_uuid)

    @staticmethod
    def generate_token_key(length: int) -> str:
        token_symbols = digits + ascii_uppercase + ascii_lowercase
        token_key = ''.join(random.choice(token_symbols) for i in range(length))  # Key length = 32 chars
        return token_key

    @staticmethod
    def get_server_setting_value(setting_name: string):
        current_setting = TeraServerSettings.get_server_setting(setting_name=setting_name)
        if current_setting:
            return current_setting.server_settings_value
        else:
            return None

    @staticmethod
    def get_server_setting(setting_name: string):
        return TeraServerSettings.query.filter_by(server_settings_name=setting_name).first()

    @staticmethod
    def set_server_setting(setting_name: string, setting_value: string):
        # Check if setting already exists
        current_setting = TeraServerSettings.get_server_setting(setting_name=setting_name)

        # Update setting if already exists
        if current_setting is not None:
            current_setting.server_settings_value = setting_value
        else:
            # Insert setting if not
            current_setting = TeraServerSettings()
            current_setting.id_server_settings = None
            current_setting.server_settings_name = setting_name
            current_setting.server_settings_value = setting_value
            TeraServerSettings.db().session.add(current_setting)
        # Store object
        current_setting.commit()

    def to_json_create_event(self):
        return self.to_json(ignore_fields=['ServerDeviceTokenKey', 'ServerParticipantTokenKey', 'ServerUUID',
                                           'ServerVersions', 'ServerDeviceRegisterKey'])

    def to_json_update_event(self):
        return self.to_json(ignore_fields=['ServerDeviceTokenKey', 'ServerParticipantTokenKey', 'ServerUUID',
                                           'ServerVersions', 'ServerDeviceRegisterKey'])



from werkzeug.local import LocalProxy
from functools import wraps
from flask import g, request
from flask_restx import reqparse
from typing import List
from enum import Enum
from flask_babel import gettext
import jwt

from opentera.services.TeraUserClient import TeraUserClient
from opentera.services.TeraDeviceClient import TeraDeviceClient
from opentera.services.TeraParticipantClient import TeraParticipantClient
from opentera.services.TeraServiceClient import TeraServiceClient
from opentera.services.DisabledTokenStorage import DisabledTokenStorage
from opentera.utils.UserAgentParser import UserAgentParser
from opentera.redis.RedisVars import RedisVars
import opentera.messages.python as messages

# Current client identity, stacked
current_user_client = LocalProxy(lambda: g.setdefault('current_user_client', None))
current_device_client = LocalProxy(lambda: g.setdefault('current_device_client', None))
current_participant_client = LocalProxy(lambda: g.setdefault('current_participant_client', None))
current_service_client = LocalProxy(lambda: g.setdefault('current_service_client', None))
current_login_type = LocalProxy(lambda: g.setdefault('current_login_type', LoginType.UNKNOWN_LOGIN))


class LoginType(Enum):
    UNKNOWN_LOGIN = 0,
    USER_LOGIN = 1,
    DEVICE_LOGIN = 2,
    PARTICIPANT_LOGIN = 3,
    SERVICE_LOGIN = 4


class ServiceAccessManager:

    api_user_token_key = None
    api_participant_token_key = None
    api_participant_static_token_key = None
    api_device_token_key = None
    api_device_static_token_key = None
    api_service_token_key = None
    token_cookie_name = 'OpenTera'
    service = None

    # Only user & participant tokens expire (for now)
    __user_disabled_token_storage = DisabledTokenStorage(redis_key='user_disabled_tokens')
    __participant_disabled_token_storage = DisabledTokenStorage(redis_key='participant_disabled_tokens')

    @staticmethod
    def init_access_manager(service):
        # Set service
        ServiceAccessManager.service = service

        # Update Service Access information
        ServiceAccessManager.api_user_token_key = \
            service.redisGet(RedisVars.RedisVar_UserTokenAPIKey)
        ServiceAccessManager.api_participant_token_key = \
            service.redisGet(RedisVars.RedisVar_ParticipantTokenAPIKey)
        ServiceAccessManager.api_participant_static_token_key = \
            service.redisGet(RedisVars.RedisVar_ParticipantStaticTokenAPIKey)
        ServiceAccessManager.api_device_token_key = \
            service.redisGet(RedisVars.RedisVar_DeviceTokenAPIKey)
        ServiceAccessManager.api_device_static_token_key = \
            service.redisGet(RedisVars.RedisVar_DeviceStaticTokenAPIKey)
        ServiceAccessManager.api_service_token_key = \
            service.redisGet(RedisVars.RedisVar_ServiceTokenAPIKey)

        # Update Token Storage information
        ServiceAccessManager.__user_disabled_token_storage.config(
            service.config_man, ServiceAccessManager.api_user_token_key)
        ServiceAccessManager.__participant_disabled_token_storage.config(
            service.config_man, ServiceAccessManager.api_participant_token_key)

    @staticmethod
    def user_add_disabled_token(token):
        return ServiceAccessManager.__user_disabled_token_storage.add_disabled_token(token)

    @staticmethod
    def is_user_token_disabled(token):
        return ServiceAccessManager.__user_disabled_token_storage.is_disabled_token(token)

    @staticmethod
    def participant_add_disabled_token(token):
        return ServiceAccessManager.__participant_disabled_token_storage.add_disabled_token(token)

    @staticmethod
    def is_participant_token_disabled(token):
        return ServiceAccessManager.__participant_disabled_token_storage.is_disabled_token(token)

    @staticmethod
    def token_required(allow_dynamic_tokens=True, allow_static_tokens=False):
        def wrap(f):
            def decorated(*args, **kwargs):
                # We support 3 authentication scheme: token in url, cookie and authorization header
                token = None
                login_infos = UserAgentParser.parse_request_for_login_infos(request)

                ######################
                # AUTHORIZATION HEADER
                if 'Authorization' in request.headers:
                    try:
                        # Default whitespace as separator, 1 split max
                        scheme, atoken = request.headers['Authorization'].split(None, 1)
                    except ValueError:
                        # malformed Authorization header
                        ServiceAccessManager.service.logger.send_login_event(
                            sender='ServiceAccessManager.token_required',
                            level=messages.LogEvent.LOGLEVEL_ERROR,
                            login_type=messages.LoginEvent.LOGIN_TYPE_TOKEN,
                            login_status=messages.LoginEvent.LOGIN_STATUS_FAILED_WITH_INVALID_TOKEN,
                            client_name=login_infos['client_name'], client_version=login_infos['client_version'],
                            client_ip=login_infos['client_ip'], os_name=login_infos['os_name'],
                            os_version=login_infos['os_version'], server_endpoint=login_infos['server_endpoint'],
                            service_uuid=ServiceAccessManager.service.service_uuid)
                        return gettext('Forbidden'), 403

                    # Verify scheme and token
                    if scheme == 'OpenTera':
                        token = atoken
                if token is None:
                    #################
                    # TOKEN PARAMETER ?
                    parser = reqparse.RequestParser()
                    parser.add_argument('token', type=str, help='Device, participant or user token', required=False)

                    # Parse arguments
                    request_args = parser.parse_args(strict=False)

                    # Verify token in params
                    if 'token' in request_args:
                        token = request_args['token']
                if token is None:
                    ###############
                    # COOKIE TOKEN
                    if ServiceAccessManager.token_cookie_name in request.cookies:
                        token = request.cookies[ServiceAccessManager.token_cookie_name]

                #########################
                # Verify token from redis
                # USER TOKEN MANAGEMENT
                if allow_dynamic_tokens:  # User only use dynamic tokens, don't validate otherwise
                    if ServiceAccessManager.validate_user_token(token=token):
                        return f(*args, **kwargs)

                # DEVICE TOKEN MANAGEMENT
                if ServiceAccessManager.validate_device_token(token=token, allow_dynamic_tokens=allow_dynamic_tokens,
                                                              allow_static_tokens=allow_static_tokens):
                    return f(*args, **kwargs)

                # PARTICIPANT TOKEN MANAGEMENT
                if ServiceAccessManager.validate_participant_token(token=token,
                                                                   allow_dynamic_tokens=allow_dynamic_tokens,
                                                                   allow_static_tokens=allow_static_tokens):
                    return f(*args, **kwargs)

                ServiceAccessManager.service.logger.send_login_event(
                    sender='ServiceAccessManager.token_required',
                    level=messages.LogEvent.LOGLEVEL_ERROR,
                    login_type=messages.LoginEvent.LOGIN_TYPE_TOKEN,
                    login_status=messages.LoginEvent.LOGIN_STATUS_FAILED_WITH_INVALID_TOKEN,
                    client_name=login_infos['client_name'], client_version=login_infos['client_version'],
                    client_ip=login_infos['client_ip'], os_name=login_infos['os_name'],
                    os_version=login_infos['os_version'], server_endpoint=login_infos['server_endpoint'],
                    service_uuid=ServiceAccessManager.service.service_uuid)
                return gettext('Forbidden'), 403

            return decorated
        return wrap

    @staticmethod
    def service_token_required(f):
        @wraps(f)
        def decorated(*args, **kwargs):
            # We support token auth. only.
            token = None
            login_infos = UserAgentParser.parse_request_for_login_infos(request)
            ######################
            # AUTHORIZATION HEADER
            if 'Authorization' in request.headers:
                try:
                    # Default whitespace as separator, 1 split max
                    scheme, atoken = request.headers['Authorization'].split(None, 1)
                except ValueError:
                    # malformed Authorization header
                    ServiceAccessManager.service.logger.send_login_event(
                        sender='ServiceAccessManager.service_token_required',
                        level=messages.LogEvent.LOGLEVEL_ERROR,
                        login_type=messages.LoginEvent.LOGIN_TYPE_TOKEN,
                        login_status=messages.LoginEvent.LOGIN_STATUS_FAILED_WITH_INVALID_TOKEN,
                        client_name=login_infos['client_name'], client_version=login_infos['client_version'],
                        client_ip=login_infos['client_ip'], os_name=login_infos['os_name'],
                        os_version=login_infos['os_version'], server_endpoint=login_infos['server_endpoint'],
                        service_uuid=ServiceAccessManager.service.service_uuid)
                    return gettext('Forbidden'), 403

                # Verify scheme and token
                if scheme == 'OpenTera':
                    token = atoken

            # Verify token from with service api key
            if ServiceAccessManager.validate_service_token(token=token):
                return f(*args, **kwargs)

            ServiceAccessManager.service.logger.send_login_event(
                sender='ServiceAccessManager.service_token_required',
                level=messages.LogEvent.LOGLEVEL_ERROR,
                login_type=messages.LoginEvent.LOGIN_TYPE_TOKEN,
                login_status=messages.LoginEvent.LOGIN_STATUS_FAILED_WITH_INVALID_TOKEN,
                client_name=login_infos['client_name'], client_version=login_infos['client_version'],
                client_ip=login_infos['client_ip'], os_name=login_infos['os_name'],
                os_version=login_infos['os_version'], server_endpoint=login_infos['server_endpoint'],
                service_uuid=ServiceAccessManager.service.service_uuid)
            return gettext('Forbidden'), 403

        return decorated

    @staticmethod
    def service_or_others_token_required(allow_dynamic_tokens=True, allow_static_tokens=False):
        def wrap(f):
            def decorated(*args, **kwargs):
                # We support 3 authentication scheme: token in url, cookie and authorization header
                token = None
                login_infos = UserAgentParser.parse_request_for_login_infos(request)
                ######################
                # AUTHORIZATION HEADER
                if 'Authorization' in request.headers:
                    try:
                        # Default whitespace as separator, 1 split max
                        scheme, atoken = request.headers['Authorization'].split(None, 1)
                    except ValueError:
                        # malformed Authorization header
                        ServiceAccessManager.service.logger.send_login_event(
                            sender='ServiceAccessManager.service_or_others_token_required',
                            level=messages.LogEvent.LOGLEVEL_ERROR,
                            login_type=messages.LoginEvent.LOGIN_TYPE_TOKEN,
                            login_status=messages.LoginEvent.LOGIN_STATUS_FAILED_WITH_INVALID_TOKEN,
                            client_name=login_infos['client_name'], client_version=login_infos['client_version'],
                            client_ip=login_infos['client_ip'], os_name=login_infos['os_name'],
                            os_version=login_infos['os_version'], server_endpoint=login_infos['server_endpoint'],
                            service_uuid=ServiceAccessManager.service.service_uuid)
                        return gettext('Forbidden'), 403

                    # Verify scheme and token
                    if scheme == 'OpenTera':
                        token = atoken

                        # Service tokens only allowed in authorization header - check here
                        if ServiceAccessManager.validate_service_token(token=token):
                            return f(*args, **kwargs)

                if token is None:
                    #################
                    # TOKEN PARAMETER ?
                    parser = reqparse.RequestParser()
                    parser.add_argument('token', type=str, help='Device, participant or user token', required=False)

                    # Parse arguments
                    request_args = parser.parse_args(strict=False)

                    # Verify token in params
                    if 'token' in request_args:
                        token = request_args['token']
                if token is None:
                    ###############
                    # COOKIE TOKEN
                    if ServiceAccessManager.token_cookie_name in request.cookies:
                        token = request.cookies[ServiceAccessManager.token_cookie_name]

                #########################
                # Verify token from redis
                # USER TOKEN MANAGEMENT
                if allow_dynamic_tokens:  # User only use dynamic tokens, don't validate otherwise
                    if ServiceAccessManager.validate_user_token(token=token):
                        return f(*args, **kwargs)

                # DEVICE TOKEN MANAGEMENT
                if ServiceAccessManager.validate_device_token(token=token, allow_dynamic_tokens=allow_dynamic_tokens,
                                                              allow_static_tokens=allow_static_tokens):
                    return f(*args, **kwargs)

                # PARTICIPANT TOKEN MANAGEMENT
                if ServiceAccessManager.validate_participant_token(token=token,
                                                                   allow_dynamic_tokens=allow_dynamic_tokens,
                                                                   allow_static_tokens=allow_static_tokens):
                    return f(*args, **kwargs)

                ServiceAccessManager.service.logger.send_login_event(
                    sender='ServiceAccessManager.service_or_others_token_required',
                    level=messages.LogEvent.LOGLEVEL_ERROR,
                    login_type=messages.LoginEvent.LOGIN_TYPE_TOKEN,
                    login_status=messages.LoginEvent.LOGIN_STATUS_FAILED_WITH_INVALID_TOKEN,
                    client_name=login_infos['client_name'], client_version=login_infos['client_version'],
                    client_ip=login_infos['client_ip'], os_name=login_infos['os_name'],
                    os_version=login_infos['os_version'], server_endpoint=login_infos['server_endpoint'],
                    service_uuid=ServiceAccessManager.service.service_uuid)
                return gettext('Forbidden'), 403

            return decorated

        return wrap

    @staticmethod
    def validate_user_token(token: str) -> bool:
        try:
            token_dict = jwt.decode(token, ServiceAccessManager.api_user_token_key, algorithms='HS256')
        except jwt.PyJWTError as e:
            # Not a user, or invalid token, will continue...
            pass
        else:
            # Check if token is disabled
            if ServiceAccessManager.is_user_token_disabled(token):
                return False

            # User token is valid and not disabled
            g.current_user_client = TeraUserClient(token_dict, token, ServiceAccessManager.service.config_man)
            g.current_login_type = LoginType.USER_LOGIN
            return True

    @staticmethod
    def validate_device_token(token: str, allow_dynamic_tokens: bool, allow_static_tokens: bool) -> bool:
        if allow_dynamic_tokens:  # Check for dynamic device token
            try:
                token_dict = jwt.decode(token, ServiceAccessManager.api_device_token_key, algorithms='HS256')
            except jwt.PyJWTError as e:
                # Not a device, or invalid token, will continue...
                pass
            else:
                # Device token
                g.current_device_client = TeraDeviceClient(token_dict, token, ServiceAccessManager.service.config_man)
                g.current_login_type = LoginType.DEVICE_LOGIN
                return True

        if allow_static_tokens:  # Check for static device token
            try:
                token_dict = jwt.decode(token, ServiceAccessManager.api_device_static_token_key, algorithms='HS256')
            except jwt.PyJWTError as e:
                # Not a device, or invalid token, will continue...
                pass
            else:
                # Device token
                g.current_device_client = TeraDeviceClient(token_dict, token, ServiceAccessManager.service.config_man)
                g.current_login_type = LoginType.DEVICE_LOGIN
                return True

        return False

    @staticmethod
    def validate_participant_token(token: str, allow_dynamic_tokens: bool, allow_static_tokens: bool) -> bool:
        if allow_dynamic_tokens:  # Check for dynamic participant token
            try:
                token_dict = jwt.decode(token, ServiceAccessManager.api_participant_token_key, algorithms='HS256')
            except jwt.PyJWTError as e:
                # Not a participant, or invalid token, will continue...
                pass
            else:
                # Look for disabled tokens, token was decoded successfully
                if ServiceAccessManager.is_participant_token_disabled(token):
                    return False

                # Participant token is not disabled, everything is ok
                g.current_participant_client = \
                    TeraParticipantClient(token_dict, token, ServiceAccessManager.service.config_man)
                g.current_login_type = LoginType.PARTICIPANT_LOGIN
                return True

        if allow_static_tokens:  # Check for static participant token
            try:
                token_dict = jwt.decode(token, ServiceAccessManager.api_participant_static_token_key,
                                        algorithms='HS256')
            except jwt.PyJWTError as e:
                # Not a participant, or invalid token, will continue...
                pass
            else:
                # Participant token
                g.current_participant_client = \
                    TeraParticipantClient(token_dict, token, ServiceAccessManager.service.config_man)
                g.current_login_type = LoginType.PARTICIPANT_LOGIN
                return True

        return False

    @staticmethod
    def validate_service_token(token: str) -> bool:
        try:
            token_dict = jwt.decode(token, ServiceAccessManager.api_service_token_key, algorithms='HS256')
        except jwt.PyJWTError as e:
            # Not a device, or invalid token, will continue...
            pass
        else:
            # Service token
            g.current_service_client = TeraServiceClient(token_dict, token, ServiceAccessManager.service.config_man)
            g.current_login_type = LoginType.SERVICE_LOGIN
            return True

        return False

    @staticmethod
    def service_user_roles_all_required(roles: List[str]):
        def wrap(f):
            @wraps(f)
            def decorated(*args, **kwargs):

                # Check if service is initialized
                if ServiceAccessManager.service is None or 'service_key' \
                        not in ServiceAccessManager.service.service_info:
                    return gettext('Forbidden'), 403

                service_key = ServiceAccessManager.service.service_info['service_key']

                # Check if user is logged in, watch out not None object but LocalProxy cannot use is None...
                if not current_user_client:
                    return gettext('Forbidden'), 403

                # Super admin pass through
                if current_user_client.user_superadmin:
                    return f(*args, **kwargs)

                # Check if user has the required role (global roles are stored in token)
                user_roles_from_token = current_user_client.get_roles_for_service(service_key)

                # Check if user has the required roles
                if not all(role in user_roles_from_token for role in roles):
                    return gettext('Forbidden'), 403

                # Everything ok, continue
                return f(*args, **kwargs)
            return decorated
        return wrap

    @staticmethod
    def service_user_roles_any_required(roles: List[str]):
        def wrap(f):
            @wraps(f)
            def decorated(*args, **kwargs):

                # Check if service is initialized
                if ServiceAccessManager.service is None or 'service_key' \
                        not in ServiceAccessManager.service.service_info:
                    return gettext('Forbidden'), 403

                service_key = ServiceAccessManager.service.service_info['service_key']

                # Check if user is logged in, watch out not None object but LocalProxy cannot use is None...
                if not current_user_client:
                    return gettext('Forbidden'), 403

                # Super admin pass through
                if current_user_client.user_superadmin:
                    return f(*args, **kwargs)

                # Check if user has the required role (global roles are stored in token)
                user_roles_from_token = current_user_client.get_roles_for_service(service_key)

                # Check if user has the required roles
                if not any(role in user_roles_from_token for role in roles):
                    return gettext('Forbidden'), 403

                # Everything ok, continue
                return f(*args, **kwargs)

            return decorated

        return wrap

    @staticmethod
    def service_user_roles_required(roles: List[str]):
        # For compatibility with old code
        return ServiceAccessManager.service_user_roles_all_required(roles)




import sys
import os
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))

from .CreateSession_pb2 import CreateSession
from .DatabaseEvent_pb2 import DatabaseEvent
from .DeviceEvent_pb2 import DeviceEvent
from .JoinSessionEvent_pb2 import JoinSessionEvent
from .ParticipantEvent_pb2 import ParticipantEvent
from .Result_pb2 import Result
from .RPCMessage_pb2 import RPCMessage
from .ServerCommand_pb2 import ServerCommand
from .StopSessionEvent_pb2 import StopSessionEvent
from .TeraEvent_pb2 import TeraEvent
from .TeraMessage_pb2 import TeraMessage
from .TeraModuleMessage_pb2 import TeraModuleMessage
from .UserEvent_pb2 import UserEvent
from .UserRegisterToEvent_pb2 import UserRegisterToEvent
from .JoinSessionReplyEvent_pb2 import JoinSessionReplyEvent
from .LeaveSessionEvent_pb2 import LeaveSessionEvent
from .LogEvent_pb2 import LogEvent
from .LeaveSessionEvent_pb2 import LeaveSessionEvent
from .LoginEvent_pb2 import LoginEvent

from google.protobuf.any_pb2 import Any




# All exported symbols
__all__ = ['CreateSession',
           'DatabaseEvent',
           'DeviceEvent',
           'JoinSessionEvent',
           'JoinSessionReplyEvent',
           'LeaveSessionEvent',
           'LogEvent',
           'ParticipantEvent',
           'Result',
           'RPCMessage',
           'ServerCommand',
           'StopSessionEvent',
           'LeaveSessionEvent',
           'TeraEvent',
           'TeraMessage',
           'TeraModuleMessage',
           'UserEvent',
           'UserRegisterToEvent',
           'LoginEvent',
           'Any']



from opentera.services.ServiceConfigManager import ServiceConfigManager
from opentera.modules.BaseModule import BaseModule

import os
import subprocess


class WebRTCModule(BaseModule):

    def __init__(self, config: ServiceConfigManager):
        BaseModule.__init__(self, config.service_config['name'] + '.WebRTCModule', config)
        self.processList = []
        self.max_sessions = self.config.webrtc_config['max_sessions']
        self.base_port = self.config.webrtc_config['local_base_port']
        self.available_ports = [port for port in range(self.base_port, self.base_port + self.max_sessions)]
        self.used_ports = []

    def setup_module_pubsub(self):
        pass

    def setup_rpc_interface(self):
        pass

    def create_webrtc_session(self, session_info):

        room_name = session_info['session_key']
        owner_uuid = session_info['session_creator_uuid']
        # make sure we kill sessions already started with this owner_uuid or room name
        self.terminate_webrtc_session_with_owner_uuid(owner_uuid)
        self.terminate_webrtc_session_with_room_name(room_name)

        # Get next available port
        port = self.get_available_port()
        key = room_name

        print(self.module_name + ' - Should create WebRTC session with name:', room_name, port, key)

        if port:
            url_users = 'https://' + self.config.webrtc_config['hostname'] + ':' \
                        + str(self.config.webrtc_config['external_port']) \
                        + '/webrtc/' + str(port) + '/users?key=' + key

            url_participants = 'https://' + self.config.webrtc_config['hostname'] + ':' \
                               + str(self.config.webrtc_config['external_port']) \
                               + '/webrtc/' + str(port) + '/participants?key=' + key

            url_devices = 'https://' + self.config.webrtc_config['hostname'] + ':' \
                          + str(self.config.webrtc_config['external_port']) \
                          + '/webrtc/' + str(port) + '/devices?key=' + key

            if self.launch_node(port=port, key=key, owner=owner_uuid, session_info=session_info):
                # Return url
                return True, {'url_users': url_users,
                              'url_participants': url_participants,
                              'url_devices': url_devices,
                              'key': key,
                              'port': port,
                              'owner': owner_uuid
                              }
            else:
                return False, {'error': 'Process not launched.'}
        else:
            return False, {'error': 'No available port left.'}

    def stop_webrtc_session(self, room_name, *args, **kwargs):
        for process_dict in self.processList:
            if process_dict['key'] == room_name:
                if self.terminate_webrtc_session_with_room_name(room_name):
                    return True, process_dict['key']
        return False, {}

    def get_webrtc_session_status(self, room_name):
        from requests import get
        for process_dict in self.processList:
            if process_dict['key'] == room_name:
                url = 'https://' + self.config.webrtc_config['hostname'] + ':' \
                      + str(self.config.webrtc_config['external_port']) \
                      + '/webrtc/' + str(process_dict['port']) + '/status?key=' + room_name
                response = get(url, timeout=5, verify=False)
                if response.status_code == 200:
                    return response.json()
        return None

    def terminate_webrtc_session_with_room_name(self, room_name):
        for process_dict in self.processList:
            if process_dict['key'] == room_name:

                self.logger.log_info(self.module_name, 'terminate_webrtc_session_with_room_name',  room_name,
                                     process_dict, 'pid', str(process_dict['process'].pid))

                # Terminate process
                process_dict['process'].kill()
                # Wait for process return
                process_dict['process'].wait()

                # Remove from process list
                self.processList.remove(process_dict)

                # Remove from used ports
                self.used_ports.remove(process_dict['port'])

                # Add to available ports
                self.available_ports.append(process_dict['port'])

                return True
        return False

    def terminate_webrtc_session_with_owner_uuid(self, owner_uuid):
        for process_dict in self.processList:
            if process_dict['owner'] == owner_uuid:

                self.logger.log_info(self.module_name, 'terminate_webrtc_session_with_owner_uuid',  owner_uuid,
                                     process_dict, 'pid', str(process_dict['process'].pid))

                # Terminate process
                process_dict['process'].kill()
                # Wait for process return
                process_dict['process'].wait()

                # Remove from process list
                self.processList.remove(process_dict)

                # Remove from used ports
                self.used_ports.remove(process_dict['port'])

                # Add to available ports
                self.available_ports.append(process_dict['port'])

                return True
        return False

    def notify_module_messages(self, pattern, channel, message):
        """
        We have received a published message from redis
        """
        print(self.module_name + ' - Received message ', pattern, channel, message)
        pass

    def get_available_port(self):
        if not self.available_ports:
            return None

        # Pop first
        port = self.available_ports.pop(0)
        self.used_ports.append(port)
        return port

    def launch_node(self, port, key, owner, session_info):
        executable_args = [self.config.webrtc_config['executable'],
                           self.config.webrtc_config['script'],
                           '--port=' + str(port),
                           '--key=' + str(key),
                           '--debug=' + str(1), 
                           '--redis_hostname=' + str(self.config.redis_config['hostname']), 
                           '--redis_port=' + str(self.config.redis_config['port']),
                           '--redis_db=' + str(self.config.redis_config['db']),
                           '--redis_username=' + str(self.config.redis_config['username']),
                           '--redis_password=' + str(self.config.redis_config['password'])]

        # stdout=os.subprocess.PIPE, stderr=os.subprocess.PIPE)
        try:
            process = subprocess.Popen(executable_args,
                                       cwd=os.path.realpath(self.config.webrtc_config['working_directory']))

            # One more process
            self.processList.append({'process': process,
                                     'port': port,
                                     'key': key,
                                     'owner': owner,
                                     'session_info': session_info})

            self.logger.log_info(self.module_name, 'launch_node', executable_args, 'pid', str(process.pid))

            print(self.module_name + ' - started process', process)
            return True
        except OSError as e:
            print(self.module_name + ' - error starting process:', e)

        return False




from opentera.forms.TeraForm import *
from flask_babel import gettext
from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess
from opentera.db.models.TeraParticipant import TeraParticipant
from opentera.db.models.TeraParticipantGroup import TeraParticipantGroup


class TeraParticipantForm:

    @staticmethod
    def get_participant_form(groups=None):
        if groups is None:
            groups = []

        form = TeraForm("participant")

        # Building lists
        # Generic list or not accessible participant (such as when creating a new one) - return all groups
        groups_list = []
        for group in groups:
            groups_list.append(TeraFormValue(value_id=group.id_participant_group, value=group.participant_group_name))

        # Sections
        section = TeraFormSection("informations", gettext("Information"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_participant", gettext("Participant ID"), "hidden", True))
        section.add_item(TeraFormItem("id_project", gettext("Projet ID"), "hidden", True))
        # section.add_item(TeraFormItem("id_site", gettext("Site ID"), "hidden", True))
        section.add_item(TeraFormItem("participant_uuid", gettext("Participant UUID"), "hidden", True))
        section.add_item(TeraFormItem("participant_name", gettext("Participant Name"), "text", True))
        section.add_item(TeraFormItem("participant_email", gettext("Participant Email"), "text", False))
        section.add_item(TeraFormItem("id_participant_group", gettext("Participant Group"), "array", False,
                                      item_values=groups_list))
        section.add_item(TeraFormItem("participant_enabled", gettext("Participant Enabled"), "boolean", True,
                                      item_default=True))
        section.add_item(TeraFormItem("participant_token_enabled", gettext("Participant Token Login Enabled"),
                                      "boolean", True, item_default=False))
        section.add_item(TeraFormItem("participant_token", gettext("Participant Token"), "label", False,
                                      item_options={"readonly": False}))
        section.add_item(TeraFormItem("participant_login_enabled", gettext("Participant Login Enabled"),
                                      "boolean", True, item_default=False))
        section.add_item(TeraFormItem("participant_username", gettext("Participant Username"), "text", True,
                                      item_condition=TeraFormItemCondition("participant_login_enabled", "=", True)))
        section.add_item(
            TeraFormItem("participant_password", gettext("Participant Password"), "password",
                         item_options={"confirm": True},
                         item_condition=TeraFormItemCondition("participant_login_enabled", "=", True)))
        section.add_item(TeraFormItem("participant_lastonline", gettext("Last Connection"), "datetime",
                                      item_options={"readonly": True}))
        section.add_item(TeraFormItem("participant_lastsession", gettext("Last Session"), "datetime",
                                      item_options={"readonly": True}))

        return form.to_dict()



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship

from enum import Enum, unique
from datetime import datetime, timedelta
import random
from sqlalchemy import asc


class TeraSessionEvent(BaseModel, SoftDeleteMixin):
    @unique
    class SessionEventTypes(Enum):
        GENERAL_ERROR = 0
        GENERAL_INFO = 1
        GENERAL_WARNING = 2
        SESSION_START = 3
        SESSION_STOP = 4
        DEVICE_ON_CHARGE = 5
        DEVICE_OFF_CHARGE = 6
        DEVICE_LOW_BATT = 7
        DEVICE_STORAGE_LOW = 8
        DEVICE_STORAGE_FULL = 9
        DEVICE_EVENT = 10
        USER_EVENT = 11
        SESSION_JOIN = 12
        SESSION_LEAVE = 13
        SESSION_JOIN_REFUSED = 14
        # Those events are generic and are usually service dependant.
        CUSTOM_EVENT1 = 100
        CUSTOM_EVENT2 = 101
        CUSTOM_EVENT3 = 102
        CUSTOM_EVENT4 = 103
        CUSTOM_EVENT5 = 104
        CUSTOM_EVENT6 = 105
        CUSTOM_EVENT7 = 106
        CUSTOM_EVENT8 = 107
        CUSTOM_EVENT9 = 108
        CUSTOM_EVENT10 = 109

    __tablename__ = 't_sessions_events'
    id_session_event = Column(Integer, Sequence('id_session_events_sequence'), primary_key=True, autoincrement=True)
    id_session = Column(Integer, ForeignKey('t_sessions.id_session', ondelete='cascade'), nullable=False)
    id_session_event_type = Column(Integer, nullable=False)
    session_event_datetime = Column(TIMESTAMP(timezone=True), nullable=False)
    session_event_text = Column(String, nullable=True)
    session_event_context = Column(String, nullable=True)

    session_event_session = relationship('TeraSession', back_populates='session_events')

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['session_event_session', 'SessionEventTypes'])
        if minimal:
            ignore_fields.extend([])

        rval = super().to_json(ignore_fields=ignore_fields)

        return rval

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraSession import TeraSession

            base_session = TeraSession.get_session_by_name('Séance #1')
            for i in range(12):
                event = TeraSessionEvent()
                event.session_event_session = base_session
                event.id_session_event_type = i
                event.session_event_datetime = datetime.now() - timedelta(hours=random.randint(0, 10)) - timedelta(
                    minutes=random.randint(0, 45))
                event.session_event_context = 'Défaut'
                event.session_event_text = str(TeraSessionEvent.SessionEventTypes(i))
                TeraSessionEvent.db().session.add(event)

            base_session = TeraSession.query.filter_by(id_creator_participant=1).first()
            for i in range(12):
                event = TeraSessionEvent()
                event.session_event_session = base_session
                event.id_session_event_type = i
                event.session_event_datetime = datetime.now() - timedelta(hours=random.randint(0, 10)) - timedelta(
                    minutes=random.randint(0, 45))
                event.session_event_context = 'Défaut'
                event.session_event_text = str(TeraSessionEvent.SessionEventTypes(i))
                TeraSessionEvent.db().session.add(event)

            TeraSessionEvent.db().session.commit()

    @staticmethod
    def get_session_event_by_id(event_id: int, with_deleted: bool = False):
        return TeraSessionEvent.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_session_event=event_id).first()

    @staticmethod
    def get_events_for_session(id_session: int, with_deleted: bool = False):
        from .TeraSession import TeraSession
        return TeraSessionEvent.db().session.query(TeraSessionEvent).execution_options(include_deleted=with_deleted)\
            .join(TeraSessionEvent.session_event_session)\
            .filter(TeraSession.id_session == id_session).order_by(asc(TeraSessionEvent.session_event_datetime)).all()



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from sqlalchemy import Column, Integer, String, Sequence, Boolean
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError

import uuid


class TeraService(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_services'

    id_service = Column(Integer, Sequence('id_service_sequence'), primary_key=True, autoincrement=True)
    service_uuid = Column(String(36), nullable=False, unique=True)
    service_name = Column(String, nullable=False)
    service_key = Column(String, nullable=False, unique=True)
    service_hostname = Column(String, nullable=False)
    service_port = Column(Integer, nullable=False)
    service_endpoint = Column(String, nullable=False)
    service_clientendpoint = Column(String, nullable=False)
    service_endpoint_user = Column(String, nullable=True)
    service_endpoint_participant = Column(String, nullable=True)
    service_endpoint_device = Column(String, nullable=True)
    service_enabled = Column(Boolean, nullable=False, default=False)
    service_system = Column(Boolean, nullable=False, default=False)
    service_editable_config = Column(Boolean, nullable=False, default=False)
    service_default_config = Column(String, nullable=True, default='{}')

    service_roles = relationship('TeraServiceRole', cascade='delete')
    service_projects = relationship('TeraServiceProject', cascade='delete')
    service_sites = relationship('TeraServiceSite', cascade='delete')

    service_created_sessions = relationship("TeraSession", cascade='delete', back_populates='session_creator_service',
                                            passive_deletes=True)

    service_assets = relationship("TeraAsset", cascade='delete', foreign_keys="TeraAsset.id_service",
                                  back_populates='asset_service', passive_deletes=True)

    service_owned_assets = relationship("TeraAsset", cascade='delete', foreign_keys="TeraAsset.asset_service_uuid",
                                        back_populates='asset_service_owner', passive_deletes=True)

    service_tests = relationship("TeraTest", cascade='delete', back_populates='test_service', passive_deletes=True)

    def __init__(self):
        pass

    def __str__(self):
        return '<TeraService ' + str(self.service_name) + ' >'

    def __repr__(self):
        return self.__str__()

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_roles', 'service_projects', 'service_sites'])

        if minimal:
            ignore_fields.extend(['service_default_config'])

        json_service = super().to_json(ignore_fields=ignore_fields)
        if not minimal:
            # Add roles for that service
            roles = []
            for role in self.service_roles:
                roles.append(role.to_json())
            json_service['service_roles'] = roles
            # Add projects for that service
            service_projects = []
            for sp in self.service_projects:
                service_projects.append(sp.to_json())
            json_service['service_projects'] = service_projects

        if not self.service_endpoint_user:
            del json_service['service_endpoint_user']

        if not self.service_endpoint_participant:
            del json_service['service_endpoint_participant']

        if not self.service_endpoint_device:
            del json_service['service_endpoint_device']

        return json_service

    def get_token(self, token_key: str, expiration=3600):
        import time
        import jwt
        # Creating token with user info
        now = time.time()
        payload = {
            'iat': int(now),
            'exp': int(now) + expiration,
            'iss': 'TeraServer',
            'service_uuid': self.service_uuid
        }

        return jwt.encode(payload, token_key, algorithm='HS256')

    @staticmethod
    def get_service_by_key(key: str, with_deleted: bool = False):
        service = TeraService.query.execution_options(include_deleted=with_deleted).filter_by(service_key=key).first()

        if service:
            return service

        return None

    @staticmethod
    def get_service_by_uuid(p_uuid: str, with_deleted: bool = False):
        service = TeraService.query.execution_options(include_deleted=with_deleted)\
            .filter_by(service_uuid=p_uuid).first()

        if service:
            return service

        return None

    @staticmethod
    def get_service_by_name(name: str, with_deleted: bool = False):
        return TeraService.query.execution_options(include_deleted=with_deleted).filter_by(service_name=name).first()

    @staticmethod
    def get_service_by_id(s_id: int, with_deleted: bool = False):
        return TeraService.query.execution_options(include_deleted=with_deleted).filter_by(id_service=s_id).first()

    @staticmethod
    def get_openteraserver_service():
        return TeraService.get_service_by_key('OpenTeraServer')

    @staticmethod
    def create_defaults(test=False):
        new_service = TeraService()
        new_service.service_uuid = '00000000-0000-0000-0000-000000000001'
        new_service.service_key = 'OpenTeraServer'
        new_service.service_name = 'OpenTera Server'
        new_service.service_hostname = '127.0.0.1'
        new_service.service_port = 4040
        new_service.service_endpoint = '/'
        new_service.service_clientendpoint = '/'
        new_service.service_endpoint_participant = '/participant'
        new_service.service_endpoint_user = '/user'
        new_service.service_endpoint_device = '/device'
        new_service.service_enabled = True
        new_service.service_system = True
        new_service.service_editable_config = True
        TeraService.db().session.add(new_service)

        new_service = TeraService()
        new_service.service_uuid = str(uuid.uuid4())
        new_service.service_key = 'LoggingService'
        new_service.service_name = 'Logging Service'
        new_service.service_hostname = '127.0.0.1'
        new_service.service_port = 4041
        new_service.service_endpoint = '/'
        new_service.service_clientendpoint = '/log'
        new_service.service_enabled = True
        new_service.service_system = True
        TeraService.db().session.add(new_service)

        new_service = TeraService()
        new_service.service_uuid = str(uuid.uuid4())
        new_service.service_key = 'FileTransferService'
        new_service.service_name = 'File Transfer Service'
        new_service.service_hostname = '127.0.0.1'
        new_service.service_port = 4042
        new_service.service_endpoint = '/'
        new_service.service_clientendpoint = '/file'
        new_service.service_enabled = True
        new_service.service_system = True
        TeraService.db().session.add(new_service)

        if test:
            new_service = TeraService()
            new_service.service_uuid = str(uuid.uuid4())
            new_service.service_key = 'BureauActif'
            new_service.service_name = 'Bureau Actif'
            new_service.service_hostname = '127.0.0.1'
            new_service.service_port = 4050
            new_service.service_endpoint = '/'
            new_service.service_clientendpoint = '/bureau'
            new_service.service_enabled = True
            TeraService.db().session.add(new_service)

        new_service = TeraService()
        new_service.service_uuid = str(uuid.uuid4())
        new_service.service_key = 'VideoRehabService'
        new_service.service_name = 'Télé-réadaptation vidéo'
        new_service.service_hostname = '127.0.0.1'
        new_service.service_port = 4070
        new_service.service_endpoint = '/'
        new_service.service_clientendpoint = '/rehab'
        new_service.service_endpoint_participant = '/participant'
        # Not yet implemented...
        # new_service.service_endpoint_user = '/user'
        # new_service.service_endpoint_device = '/device'
        new_service.service_enabled = True
        new_service.service_editable_config = True
        new_service.service_system = True
        TeraService.db().session.add(new_service)

        if test:
            new_service = TeraService()
            new_service.service_uuid = str(uuid.uuid4())
            new_service.service_key = 'RobotTeleOperationService'
            new_service.service_name = 'Robot Teleoperation Service'
            new_service.service_hostname = '127.0.0.1'
            new_service.service_port = 4080
            new_service.service_endpoint = '/'
            new_service.service_clientendpoint = '/robot'
            new_service.service_enabled = True
            TeraService.db().session.add(new_service)
            TeraService.db().session.commit()

    def to_json_create_event(self):
        return self.to_json(minimal=False)

    def to_json_update_event(self):
        return self.to_json(minimal=False)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_service': self.id_service, 'service_key': self.service_key}

    @classmethod
    def insert(cls, service):
        service.service_uuid = str(uuid.uuid4())
        super().insert(service)

        # Create default admin-user role for each service
        # from opentera.db.models.TeraServiceRole import TeraServiceRole
        # new_role = TeraServiceRole()
        # new_role.id_service = service.id_service
        # new_role.service_role_name = 'admin'
        # TeraServiceRole.insert(new_role)
        #
        # new_role = TeraServiceRole()
        # new_role.id_service = service.id_service
        # new_role.service_role_name = 'user'
        # TeraServiceRole.insert(new_role)

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        for service_site in self.service_sites:
            if service_site.delete_check_integrity(with_deleted=with_deleted):
                return IntegrityError('Have sessions, assets or tests using that service', self.id_service,
                                      't_sessions')
        return None

    @classmethod
    def update(cls, update_id: int, values: dict):
        # Prevent changes on UUID
        if 'service_uuid' in values:
            del values['service_uuid']

        super().update(update_id, values)

    def get_undelete_cascade_relations(self) -> list:
        return ['service_sites', 'service_projects', 'service_roles']



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: TeraModuleMessage.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17TeraModuleMessage.proto\x12\x11opentera.protobuf\x1a\x19google/protobuf/any.proto\"\xc6\x01\n\x11TeraModuleMessage\x12\x39\n\x04head\x18\x01 \x01(\x0b\x32+.opentera.protobuf.TeraModuleMessage.Header\x12\"\n\x04\x64\x61ta\x18\x02 \x03(\x0b\x32\x14.google.protobuf.Any\x1aR\n\x06Header\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x0c\n\x04time\x18\x02 \x01(\x01\x12\x0b\n\x03seq\x18\x03 \x01(\r\x12\x0e\n\x06source\x18\x04 \x01(\t\x12\x0c\n\x04\x64\x65st\x18\x05 \x01(\tb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'TeraModuleMessage_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TERAMODULEMESSAGE._serialized_start=74
  _TERAMODULEMESSAGE._serialized_end=272
  _TERAMODULEMESSAGE_HEADER._serialized_start=190
  _TERAMODULEMESSAGE_HEADER._serialized_end=272
# @@protoc_insertion_point(module_scope)



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from sqlalchemy import Column, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError
from opentera.db.models.TeraSite import TeraSite
from opentera.db.models.TeraProject import TeraProject
from opentera.db.models.TeraSessionUsers import TeraSessionUsers
from opentera.db.models.TeraSession import TeraSession
from opentera.db.models.TeraTest import TeraTest
from opentera.db.models.TeraAsset import TeraAsset
from opentera.db.models.TeraServiceRole import TeraServiceRole


from passlib.hash import bcrypt
import uuid
import datetime
import json
import time
import jwt


# Generator for jti
def infinite_jti_sequence():
    num = 0
    while True:
        yield num
        num += 1


# Initialize generator, call next(user_jti_generator) to get next sequence number
user_jti_generator = infinite_jti_sequence()


class TeraUser(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_users'
    id_user = Column(Integer, Sequence('id_user_sequence'), primary_key=True, autoincrement=True)
    user_username = Column(String(50), nullable=False, unique=True)
    user_uuid = Column(String(36), nullable=False, unique=True)
    user_email = Column(String, nullable=True)
    user_firstname = Column(String, nullable=False)
    user_lastname = Column(String, nullable=False)
    user_password = Column(String, nullable=False)
    user_enabled = Column(Boolean, nullable=False)
    user_profile = Column(String, nullable=False)  # Used to store "extra" informations, if needed.
    user_notes = Column(String, nullable=True)
    user_lastonline = Column(TIMESTAMP(timezone=True), nullable=True)
    user_superadmin = Column(Boolean, nullable=False, default=False)

    # user_sites_access = relationship('TeraSiteAccess', cascade="all,delete")
    # user_projects_access = relationship("TeraProjectAccess", cascade="all,delete")
    user_user_groups = relationship("TeraUserGroup", secondary="t_users_users_groups",
                                    back_populates="user_group_users", passive_deletes=True)
    user_sessions = relationship("TeraSession", secondary="t_sessions_users", back_populates="session_users",
                                 passive_deletes=True, cascade='delete')

    user_created_sessions = relationship("TeraSession", cascade='delete', back_populates='session_creator_user',
                                         passive_deletes=True)

    user_service_config = relationship("TeraServiceConfig", cascade='delete', passive_deletes=True)

    user_assets = relationship("TeraAsset", cascade='delete', back_populates='asset_user', passive_deletes=True)
    user_tests = relationship("TeraTest", cascade='delete', back_populates='test_user', passive_deletes=True)

    authenticated = False

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []
        ignore_fields.extend(['authenticated', 'user_password', 'user_user_groups',
                              'user_sessions'])
        if minimal:
            ignore_fields.extend(['user_username', 'user_email', 'user_profile', 'user_notes', 'user_lastonline',
                                  'user_superadmin'])
        rval = super().to_json(ignore_fields=ignore_fields)
        rval['user_name'] = self.get_fullname()
        return rval

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_user': self.id_user, 'user_uuid': self.user_uuid}

    def get_token(self, token_key: str, expiration: int = 3600):
        """
        Generates a token for the user. The token will contain the following information:
        - issue time (iat)
        - expiration time (exp)
        - issuer (iss)
        - token id (jti)
        - user_uuid (string, user's uuid)
        - id_user (int, user's id)
        - user_fullname (string, user's fullname)
        - user_superadmin (boolean, true if user is superadmin)
        - service_access dict containing all global services access in a dict of the form:
            {'service_access': {'service_key': ['access1', 'access2']}}

        :param token_key: Key to use to generate the token
        :param expiration: Expiration time in seconds
        """

        # Creating token with user info
        now = time.time()

        payload = {
            'iat': int(now),
            'exp': int(now) + expiration,
            'iss': 'TeraServer',
            'jti': next(user_jti_generator),
            'user_uuid': self.user_uuid,
            'id_user': self.id_user,
            'user_fullname': self.get_fullname(),
            'user_superadmin': self.user_superadmin,
            'service_access': self.get_service_access_dict()
        }

        return jwt.encode(payload, token_key, algorithm='HS256')

    def get_service_access_dict(self):
        service_access = {}

        # Service access are defined in user groups, not needed for superadmin
        if not self.user_superadmin:
            for user_group in self.user_user_groups:
                for service_role in user_group.user_group_services_roles:
                    service = service_role.service_role_service
                    role_name = service_role.service_role_name
                    service_key = service.service_key

                    if service_role.id_site is None and service_role.id_project is None:
                        # Global access
                        # Create entry if not exists
                        if service_key not in service_access:
                            service_access[service_key] = []
                        # Add role to service
                        service_access[service_key].append(role_name)

        return service_access

    def get_fullname(self):
        return self.user_firstname + ' ' + self.user_lastname

    def is_authenticated(self):
        return self.authenticated

    def is_anonymous(self):
        return False

    def is_active(self):
        return self.user_enabled

    def get_id(self):
        return self.user_uuid

    def update_last_online(self):
        self.user_lastonline = datetime.datetime.now()
        TeraUser.db().session.commit()

    def __str__(self):
        return '<TeraUser ' + str(self.user_username) + ', ' + str(self.user_email) + ' >'

    def __repr__(self):
        return self.__str__()

    def get_sites_roles(self, id_service: int | None = None) -> dict:
        sites_roles = {}

        if self.user_superadmin:
            # Super admin - admin role in all sites
            if not id_service:
                sites = TeraSite.query.all()
            else:
                all_roles = TeraServiceRole.get_service_roles(service_id=id_service)
                sites = []
                for role in all_roles:
                    if role.id_site:
                        sites.append(role.service_role_site)

            for site in sites:
                sites_roles[site] = {'site_role': 'admin', 'inherited': True}
            return sites_roles

        # Browse all user groups to get roles for those sites
        for user_group in self.user_user_groups:
            user_group_roles = user_group.get_sites_roles(service_id=id_service)
            for site, site_role in user_group_roles.items():
                if site not in sites_roles:
                    # Site not already present
                    sites_roles[site] = site_role
                else:
                    # Site present - check if we have an "admin" role to overwrite an "user" role
                    if site_role['site_role'] == 'admin':
                        sites_roles[site] = site_role

        return sites_roles

    def get_projects_roles(self, id_service: int | None = None) -> dict:
        projects_roles = {}

        if self.user_superadmin:
            # Super admin - admin role in all projects
            if not id_service:
                projects = TeraProject.query.all()
            else:
                all_roles = TeraServiceRole.get_service_roles(service_id=id_service)
                projects = []
                for role in all_roles:
                    if role.id_project:
                        projects.append(role.service_role_project)

            for project in projects:
                projects_roles[project] = {'project_role': 'admin', 'inherited': True}
            return projects_roles

        # Browse all user groups to get roles for those projects
        for user_group in self.user_user_groups:
            user_group_roles = user_group.get_projects_roles(service_id=id_service)
            for project, project_role in user_group_roles.items():
                if project not in projects_roles:
                    # Project not already present
                    projects_roles[project] = project_role
                else:
                    # Project present - check if we have an "admin" role to overwrite an "user" role
                    if project_role['project_role'] == 'admin':
                        projects_roles[project] = project_role

        return projects_roles

    def get_service_roles(self, id_service: int) -> list:
        service_roles = []

        if self.user_superadmin:
            # Super admin - has all service roles
            all_roles = TeraServiceRole.get_service_roles(service_id=id_service, globals_only=True)
            for role in all_roles:
                service_roles.append(role.service_role_name)
        else:
            # Browse all user groups to get roles
            for user_group in self.user_user_groups:
                user_group_roles = user_group.get_global_roles(service_id=id_service)
                for role in user_group_roles:
                    if role not in service_roles:
                        service_roles.append(role)

        return service_roles

    @staticmethod
    def encrypt_password(password):
        return bcrypt.hash(password)

    @staticmethod
    def verify_password(username, password, user=None):
        # Query User with that username
        if user is None:
            user = TeraUser.get_user_by_username(username)
        if user is None:
            print('TeraUser: verify_password - user ' + username + ' not found.')
            return None

        # Check if enabled
        if not user.user_enabled:
            print('TeraUser: verify_password - user ' + username + ' is disabled.')
            return None

        # Check password
        if bcrypt.verify(password, user.user_password):
            user.authenticated = True
            return user

        return None

    @staticmethod
    def get_user_by_username(username, with_deleted: bool = False):
        return TeraUser.query.execution_options(include_deleted=with_deleted).filter_by(user_username=username).first()

    @staticmethod
    def get_user_by_uuid(u_uuid, with_deleted: bool = False):
        user = TeraUser.query.execution_options(include_deleted=with_deleted).filter_by(user_uuid=u_uuid).first()
        return user

    @staticmethod
    def get_user_by_id(id_user, with_deleted: bool = False):
        user = TeraUser.query.execution_options(include_deleted=with_deleted).filter_by(id_user=id_user).first()
        return user

    @staticmethod
    def get_superadmins(with_deleted: bool = False):
        return TeraUser.query.execution_options(include_deleted=with_deleted).filter_by(user_superadmin=True).all()

    @classmethod
    def update(cls, id_user: int, values: dict):
        # Remove the password field is present and if empty
        if 'user_password' in values:
            if values['user_password'] == '':
                del values['user_password']
            else:
                # Forcing password to string
                values['user_password'] = TeraUser.encrypt_password(str(values['user_password']))

        # Dumps the user profile if present
        if 'user_profile' in values:
            if isinstance(values['user_profile'], dict):
                values['user_profile'] = json.dumps(values['user_profile'])

        # Prevent changes on UUID
        if 'user_uuid' in values:
            del values['user_uuid']

        super().update(id_user, values)

    @classmethod
    def insert(cls, user):
        # Encrypts password
        # Forcing password to string
        user.user_password = TeraUser.encrypt_password(str(user.user_password))

        # Generate UUID
        user.user_uuid = str(uuid.uuid4())

        # Clear last online field
        user.user_lastonline = None

        # Converts profile from dict, if needed
        if isinstance(user.user_profile, dict):
            user.user_profile = json.dumps(user.user_profile)

        super().insert(user)

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        if TeraSessionUsers.get_session_count_for_user(self.id_user, with_deleted=with_deleted) > 0:
            return IntegrityError('User still has sessions', self.id_user, 't_sessions_users')

        if TeraSession.get_count(filters={'id_creator_user': self.id_user}, with_deleted=with_deleted) > 0:
            return IntegrityError('User still has created sessions', self.id_user, 't_sessions')

        if TeraAsset.get_count(filters={'id_user': self.id_user}, with_deleted=with_deleted) > 0:
            return IntegrityError('User still has created assets', self.id_user, 't_assets')

        if TeraTest.get_count(filters={'id_user': self.id_user}, with_deleted=with_deleted) > 0:
            return IntegrityError('User still has created tests', self.id_user, 't_tests')

        return None

    @staticmethod
    def create_defaults(test=False):
        # Admin
        admin = TeraUser()
        admin.user_enabled = True
        admin.user_firstname = "Super"
        admin.user_lastname = "Admin"
        admin.user_profile = ""
        admin.user_password = TeraUser.encrypt_password("admin")
        admin.user_superadmin = True
        admin.user_username = "admin"
        admin.user_uuid = str(uuid.uuid4())
        TeraUser.db().session.add(admin)

        if test:
            # Site admin
            admin = TeraUser()
            admin.user_enabled = True
            admin.user_firstname = "Site"
            admin.user_lastname = "Admin"
            admin.user_profile = ""
            admin.user_password = TeraUser.encrypt_password("siteadmin")
            admin.user_superadmin = False
            admin.user_username = "siteadmin"
            admin.user_uuid = str(uuid.uuid4())
            # admin.user_user_group = TeraUserGroup.get_user_group_by_group_name("Admins - Default Site")
            TeraUser.db().session.add(admin)

            # Site User
            user = TeraUser()
            user.user_enabled = True
            user.user_firstname = "Site"
            user.user_lastname = "User"
            user.user_profile = ""
            user.user_password = TeraUser.encrypt_password("user")
            user.user_superadmin = False
            user.user_username = "user"
            user.user_uuid = str(uuid.uuid4())
            # user.user_user_group = TeraUserGroup.get_user_group_by_group_name("Users - Project 1")
            TeraUser.db().session.add(user)

            # Site User
            user = TeraUser()
            user.user_enabled = True
            user.user_firstname = "MultiSite"
            user.user_lastname = "User"
            user.user_profile = ""
            user.user_password = TeraUser.encrypt_password("user2")
            user.user_superadmin = False
            user.user_username = "user2"
            user.user_uuid = str(uuid.uuid4())
            # user.user_user_group = TeraUserGroup.get_user_group_by_group_name("Users - Projects 1 & 2")
            TeraUser.db().session.add(user)

            # Project admin
            user = TeraUser()
            user.user_enabled = True
            user.user_firstname = "Project"
            user.user_lastname = "Admin"
            user.user_profile = ""
            user.user_password = TeraUser.encrypt_password("user3")
            user.user_superadmin = False
            user.user_username = "user3"
            user.user_uuid = str(uuid.uuid4())
            # user.user_user_group = TeraUserGroup.get_user_group_by_group_name("Users - Projects 1 & 2")
            TeraUser.db().session.add(user)

            # No access user!
            user = TeraUser()
            user.user_enabled = True
            user.user_firstname = "No Access"
            user.user_lastname = "User!"
            user.user_profile = ""
            user.user_password = TeraUser.encrypt_password("user4")
            user.user_superadmin = False
            user.user_username = "user4"
            user.user_uuid = str(uuid.uuid4())
            # user.user_user_group = TeraUserGroup.get_user_group_by_group_name("Users - Projects 1 & 2")
            TeraUser.db().session.add(user)

        TeraUser.db().session.commit()

    def get_undelete_cascade_relations(self) -> list:
        return ['user_service_config']



# This file is automatically generated by cmake. Do not change.
opentera_server_version_string = "OpenTera_v1.2.5"
opentera_server_major_version = 1
opentera_server_minor_version = 2
opentera_server_patch_version = 5



from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives.serialization import (Encoding, PrivateFormat, NoEncryption)
import datetime
import os
import uuid
import socket


def load_private_pem_key(filename):
    with open(filename, 'rb') as key:
        private_key = serialization.load_pem_private_key(key.read(), None, default_backend())
        return private_key

    return None


def load_pem_certificate(filename):
    with open(filename, 'rb') as f:
        cert = x509.load_pem_x509_certificate(f.read(), default_backend())
        return cert

    return None


def load_pem_certificate_from_data(data: bytes):
    return x509.load_pem_x509_certificate(data, default_backend())


# info at https://cryptography.io
def generate_ca_certificate(common_name=socket.gethostname(), country_name=u'CA',
                            state_or_province=u'State_Or_Province', locality_name=u'Locality Name',
                            organization_name=u'Organization Name', email=u'test@noemail.com'):

    result = {}

    # Generate the private key
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend())

    result['private_key'] = private_key

    subject = issuer = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        x509.NameAttribute(NameOID.COUNTRY_NAME, country_name),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, state_or_province),
        x509.NameAttribute(NameOID.LOCALITY_NAME, locality_name),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, organization_name),
        x509.NameAttribute(NameOID.EMAIL_ADDRESS, email)
    ])

    builder = x509.CertificateSigningRequestBuilder()

    # Create Certificate
    cert = x509.CertificateBuilder().subject_name(
            subject
        ).issuer_name(
            issuer
        ).public_key(
            private_key.public_key()
        ).serial_number(
            x509.random_serial_number()
        ).not_valid_before(
            datetime.datetime.utcnow() - datetime.timedelta(days=1)
        ).not_valid_after(
            # Our certificate will be valid for 10 years
            datetime.datetime.utcnow() + datetime.timedelta(days=3650)
        ).add_extension(
            x509.BasicConstraints(ca=True, path_length=None),
            critical=True
            # Sign our certificate with our private key
        ).sign(private_key, hashes.SHA256(), default_backend())

    result['certificate'] = cert

    return result


def generate_local_certificate_csr(common_name=socket.gethostname(), country_name=u'CA',
                                   state_or_province=u'State_Or_Province', locality_name=u'Locality Name',
                                   organization_name=u'Organization Name', email=u'test@noemail.com'):

    result = {}

    # 1. You generate a private / public key pair
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())

    result['private_key'] = private_key

    # 2. You create a request for a certificate, which is signed by your key (to prove that you own the key)
    csr = x509.CertificateSigningRequestBuilder().subject_name(x509.Name([
        # Provide various details about who we are.
        x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        x509.NameAttribute(NameOID.COUNTRY_NAME, country_name),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, state_or_province),
        x509.NameAttribute(NameOID.LOCALITY_NAME, locality_name),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, organization_name),
        x509.NameAttribute(NameOID.EMAIL_ADDRESS, email)
    ])).sign(private_key, hashes.SHA256(), default_backend())

    result['csr'] = csr

    return result


def generate_local_certificate(common_name=socket.gethostname(), country_name=u'CA',
                               state_or_province=u'State_Or_Province', locality_name=u'Locality Name',
                               organization_name=u'Organization Name', email=u'test@noemail.com'):

    result = {}

    # Generate the private key
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend())

    result['private_key'] = private_key

    subject = issuer = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        x509.NameAttribute(NameOID.COUNTRY_NAME, country_name),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, state_or_province),
        x509.NameAttribute(NameOID.LOCALITY_NAME, locality_name),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, organization_name),
        x509.NameAttribute(NameOID.EMAIL_ADDRESS, email)
    ])

    builder = x509.CertificateSigningRequestBuilder()

    # Create Certificate
    cert = x509.CertificateBuilder().subject_name(
            subject
        ).issuer_name(
            issuer
        ).public_key(
            private_key.public_key()
        ).serial_number(
            x509.random_serial_number()
        ).not_valid_before(
            datetime.datetime.utcnow() - datetime.timedelta(hours=1)
        ).not_valid_after(
            # Our certificate will be valid for 1 year
            datetime.datetime.utcnow() + datetime.timedelta(days=365)
        ).add_extension(
            x509.SubjectAlternativeName([x509.DNSName(u"localhost")]),
            critical=False
            # Sign our certificate with our private key
        ).add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True
        ).sign(private_key, hashes.SHA256(), default_backend())

    result['certificate'] = cert

    return result


def write_private_key_and_certificate(info: dict, keyfile='key.pem', certfile='cert.pem'):
    """
        Use def generate_ca_certificate(...) to generate the private key and certificate before.

    """
    try:
        # Will write private key in PEM (base64) format
        with open(keyfile, 'wb') as f:
            f.write(info['private_key'].private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=NoEncryption()
            ))

        # Will write certificate
        with open(certfile, 'wb') as f:
            f.write(info['certificate'].public_bytes(serialization.Encoding.PEM))
    except:
        return False

    return True


# Default apple watch, for testing...
def create_certificate_signing_request(device_name):

    result = {}

    # 1. You generate a private / public key pair
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())

    result['private_key'] = private_key

    # 2. You create a request for a certificate, which is signed by your key (to prove that you own the key)
    csr = x509.CertificateSigningRequestBuilder().subject_name(x509.Name([
        # Provide various details about who we are.
        x509.NameAttribute(NameOID.COMMON_NAME, str(device_name))
        ])).sign(private_key, hashes.SHA256(), default_backend())

    result['csr'] = csr

    return result


def generate_local_certificate_with_csr(csr, ca_info):
    # 3 You give your csr to a CA (but not the private key!)
    # 4 The CA validates that you own the resource (verify uuid?)
    # 5 The CA gives you a certificate, signed by them, which identifies your public key, and the resource you are
    #   authenticated for.
    builder = x509.CertificateBuilder()

    ca = ca_info['certificate']
    ca_key = ca_info['private_key']

    # # WARNING, subject needs to be verified by CA, do not use csr subject
    builder = builder.subject_name(
        x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, u'127.0.0.1'),
            x509.NameAttribute(NameOID.COUNTRY_NAME, u'CA'),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u'State'),
            x509.NameAttribute(NameOID.LOCALITY_NAME, u'Locality'),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, u'Organization'),
            x509.NameAttribute(NameOID.EMAIL_ADDRESS, u'test@noemail.com')
        ])
    )

    builder = builder.add_extension(
        x509.SubjectAlternativeName([x509.DNSName(u'localhost')]),
        critical=False
    )

    # Certificate valid for one year
    builder = builder.issuer_name(ca.subject)
    builder = builder.not_valid_before(datetime.datetime.now() - datetime.timedelta(hours=1))
    builder = builder.not_valid_after(datetime.datetime.now() + datetime.timedelta(days=365))
    builder = builder.public_key(csr.public_key())
    builder = builder.serial_number(x509.random_serial_number())
    for ext in csr.extensions:
        builder = builder.add_extension(ext.value, ext.critical)

    # Sign with the CA
    certificate = builder.sign(
        private_key=ca_key,
        algorithm=hashes.SHA256(),
        backend=default_backend()
    )

    return certificate


def generate_device_certificate(csr, ca_info, device_uuid):
    # 3 You give your csr to a CA (but not the private key!)
    # 4 The CA validates that you own the resource (verify uuid?)
    # 5 The CA gives you a certificate, signed by them, which identifies your public key, and the resource you are
    #   authenticated for.
    builder = x509.CertificateBuilder()

    ca = ca_info['certificate']
    ca_key = ca_info['private_key']

    builder = builder.subject_name(
        x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, u'Device'),
            x509.NameAttribute(NameOID.USER_ID, str(device_uuid))
        ])
    )

    builder = builder.issuer_name(ca.subject)
    builder = builder.not_valid_before(datetime.datetime.now() - datetime.timedelta(hours=1))
    builder = builder.not_valid_after(datetime.datetime.now() + datetime.timedelta(days=3650))
    builder = builder.public_key(csr.public_key())
    builder = builder.serial_number(x509.random_serial_number())
    for ext in csr.extensions:
        builder = builder.add_extension(ext.value, ext.critical)

    # Sign with the CA
    certificate = builder.sign(
        private_key=ca_key,
        algorithm=hashes.SHA256(),
        backend=default_backend()
    )

    return certificate


# For testing...
if __name__ == '__main__':
    print(socket.gethostname())

    current_path = os.getcwd()

    # Generate CA certificate
    ca_info = generate_ca_certificate(common_name='IntRoLab_CA')
    write_private_key_and_certificate(ca_info, keyfile=current_path + '/../../certificates/ca_key.pem',
                                      certfile=current_path + '/../../certificates/ca_cert.pem')

    # Generate Local Certificate key and csr
    local_info = generate_local_certificate_csr()
    print(local_info)

    # Create local certificate signed by our CA
    local_info['certificate'] = generate_local_certificate_with_csr(local_info['csr'], ca_info)
    print(local_info)

    write_private_key_and_certificate(local_info, keyfile=current_path + '/../../certificates/site_key.pem',
                                      certfile=current_path + '/../../certificates/site_cert.pem')

    # # Generate signing request
    # client_info = create_certificate_signing_request(device_name='Test Device')
    # client_info['certificate'] = generate_device_certificate(client_info['csr'], ca_info,
    #                                                          'b707e0b2-e649-47e7-a938-2b949c423f73')
    #
    # # Client test key + certificate
    # write_private_key_and_certificate(client_info,
    #                                   keyfile=current_path + '/../../certificates/devices/client_key.pem',
    #                                   certfile=current_path + '/../../certificates/devices/client_certificate.pem')
    #
    # # Will use hostname as common name by default
    # site_info = generate_local_certificate()
    # write_private_key_and_certificate(site_info, keyfile=current_path + '/../../certificates/site_key.pem',
    #                                   certfile=current_path + '/../../certificates/site_cert.pem')
    #
    # print(ca_info)
    # print(site_info)
    # print(client_info)



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship


class TeraSessionDevices(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_sessions_devices'
    id_session_device = Column(Integer, Sequence('id_session_device'), primary_key=True, autoincrement=True)
    id_session = Column(Integer, ForeignKey('t_sessions.id_session', ondelete='cascade'))
    id_device = Column(Integer, ForeignKey('t_devices.id_device'))

    session_device_session = relationship('TeraSession', viewonly=True)
    session_device_device = relationship('TeraDevice', viewonly=True)

    @classmethod
    def update(cls, update_id: int, values: dict):
        return



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.models.TeraDeviceType import TeraDeviceType
from opentera.db.models.TeraServerSettings import TeraServerSettings
from opentera.db.models.TeraSessionDevices import TeraSessionDevices
from opentera.db.models.TeraAsset import TeraAsset
from opentera.db.models.TeraTest import TeraTest
from opentera.db.models.TeraSession import TeraSession
from opentera.db.models.TeraDeviceParticipant import TeraDeviceParticipant

import uuid
import jwt
import datetime
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraDevice(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_devices'
    id_device = Column(Integer, Sequence('id_device_sequence'), primary_key=True, autoincrement=True)

    device_uuid = Column(String(36), nullable=False, unique=True)
    device_name = Column(String, nullable=False)
    id_device_type = Column(Integer, ForeignKey('t_devices_types.id_device_type', ondelete='cascade'), nullable=False)
    id_device_subtype = Column(Integer, ForeignKey('t_devices_subtypes.id_device_subtype',
                                                   ondelete='set null'), nullable=True)
    device_token = Column(String, nullable=True, unique=True)
    device_certificate = Column(String, nullable=True)
    device_enabled = Column(Boolean, nullable=False, default=False)
    device_onlineable = Column(Boolean, nullable=False, default=False)
    device_config = Column(String, nullable=True)
    device_infos = Column(String, nullable=True)
    device_notes = Column(String, nullable=True)
    device_lastonline = Column(TIMESTAMP(timezone=True), nullable=True)

    device_sites = relationship("TeraSite", secondary='t_devices_sites', back_populates='site_devices')
    # device_projects = relationship('TeraDeviceProject', cascade='delete')
    device_projects = relationship("TeraProject", secondary="t_devices_projects",
                                   back_populates="project_devices", lazy='selectin')
    # device_session_types = relationship("TeraSessionTypeDeviceType")
    device_participants = relationship("TeraParticipant",  secondary="t_devices_participants",
                                       back_populates="participant_devices", passive_deletes=True)
    device_sessions = relationship("TeraSession", secondary="t_sessions_devices", back_populates="session_devices",
                                   passive_deletes=True)
    device_type = relationship('TeraDeviceType')
    device_subtype = relationship('TeraDeviceSubType')
    device_created_sessions = relationship("TeraSession", cascade='delete', back_populates='session_creator_device',
                                           passive_deletes=True)

    device_service_config = relationship("TeraServiceConfig", cascade='delete', passive_deletes=True)

    device_assets = relationship("TeraAsset", cascade='delete', back_populates='asset_device', passive_deletes=True)

    device_tests = relationship("TeraTest", cascade='delete', back_populates='test_device', passive_deletes=True)

    authenticated = False

    # def __init__(self):
    #     self.secret = TeraServerSettings.get_server_setting_value(TeraServerSettings.ServerDeviceTokenKey)
    #     if self.secret is None:
    #         # Fallback - should not happen
    #         self.secret = 'TeraDeviceSecret'

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields += ['device_projects', 'device_participants', 'device_sessions', 'device_certificate',
                          'device_type', 'device_subtype', 'authenticated', 'device_assets', 'device_sites']

        if minimal:
            ignore_fields += ['device_onlineable', 'device_config', 'device_notes',
                              'device_lastonline', 'device_infos',  'device_token']

        device_json = super().to_json(ignore_fields=ignore_fields)

        return device_json

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_device': self.id_device, 'device_uuid': self.device_uuid}

    def from_json(self, json):
        super().from_json(json)

        # Manage device subtype
        if type(self.device_subtype) is dict:
            self.id_device_subtype = self.device_subtype['id_device_subtype']
            self.device_subtype = None

    def is_authenticated(self):
        return self.authenticated

    def is_anonymous(self):
        return False

    def is_active(self):
        return self.device_enabled

    def get_id(self):
        return self.device_uuid

    def is_login_enabled(self):
        return self.device_onlineable

    def create_token(self):
        # Creating token with user info
        payload = {
            'iss': 'TeraServer',
            'device_uuid': self.device_uuid,
            'id_device': self.id_device
        }

        self.device_token = jwt.encode(payload, TeraServerSettings.get_server_setting_value(
            TeraServerSettings.ServerDeviceTokenKey), algorithm='HS256')

        return self.device_token

    def update_last_online(self):
        self.device_lastonline = datetime.datetime.now()
        TeraDevice.db().session.commit()

    @staticmethod
    def get_device_by_token(token, with_deleted: bool = False):
        device = TeraDevice.query.execution_options(include_deleted=with_deleted).filter_by(device_token=token).first()

        if device:
            # Validate token, key loaded from DB
            data = jwt.decode(token.encode('utf-8'),
                              TeraServerSettings.get_server_setting_value(
                                  TeraServerSettings.ServerDeviceTokenKey), algorithms='HS256')

            # Only validating UUID since other fields can change in database after token is generated.
            if data['device_uuid'] == device.device_uuid:
                device.authenticated = True
                return device
            else:
                return None

        return None

    @staticmethod
    def get_device_by_certificate(certificate, with_deleted: bool = False):
        return TeraDevice.query.execution_options(include_deleted=with_deleted).\
            filter_by(device_certificate=certificate).first()

    @staticmethod
    def get_device_by_uuid(dev_uuid, with_deleted: bool = False):
        device = TeraDevice.query.execution_options(include_deleted=with_deleted).\
            filter_by(device_uuid=dev_uuid).first()
        return device

    @staticmethod
    def get_device_by_name(name, with_deleted: bool = False):
        return TeraDevice.query.execution_options(include_deleted=with_deleted).filter_by(device_name=name).first()

    @staticmethod
    def get_device_by_id(device_id, with_deleted: bool = False):
        return TeraDevice.query.execution_options(include_deleted=with_deleted).filter_by(id_device=device_id).first()

    @staticmethod
    def create_defaults(test=False):
        if test:
            device = TeraDevice()
            device.device_name = 'Apple Watch #W05P1'
            # Forcing uuid for tests
            device.device_uuid = 'b707e0b2-e649-47e7-a938-2b949c423f73'  # str(uuid.uuid4())
            device.id_device_type = TeraDeviceType.get_device_type_by_key('capteur').id_device_type
            # device.create_token()
            device.device_enabled = True
            device.device_onlineable = True
            # device.device_site = TeraSite.get_site_by_sitename('Default Site')
            # device.device_participants = [TeraParticipant.get_participant_by_id(1)]
            TeraDevice.db().session.add(device)

            device2 = TeraDevice()
            device2.device_name = 'Kit Télé #1'
            device2.device_uuid = str(uuid.uuid4())
            device2.id_device_type = TeraDeviceType.get_device_type_by_key('videoconference').id_device_type
            # device2.create_token()
            device2.device_enabled = True
            device2.device_onlineable = True
            # device2.device_sites = [TeraSite.get_site_by_sitename('Default Site')]
            # device2.device_participants = [TeraParticipant.get_participant_by_id(1),
            #                               TeraParticipant.get_participant_by_id(2)]
            TeraDevice.db().session.add(device2)

            device3 = TeraDevice()
            device3.device_name = 'Robot A'
            device3.device_uuid = str(uuid.uuid4())
            device3.id_device_type = TeraDeviceType.get_device_type_by_key('robot').id_device_type
            # device3.create_token()
            device3.device_enabled = False
            device3.device_onlineable = True
            # device3.device_sites = [TeraSite.get_site_by_sitename('Default Site')]
            # device3.device_participants = [TeraParticipant.get_participant_by_id(2)]
            TeraDevice.db().session.add(device3)

            TeraDevice.db().session.commit()

            # Must create token after devices are created since token contains id_device!
            device.create_token()
            device2.create_token()
            device3.create_token()

            TeraDevice.db().session.commit()

    @classmethod
    def insert(cls, device):
        # Generate UUID
        if not device.device_uuid:
            device.device_uuid = str(uuid.uuid4())

        # Clear last online field
        device.device_lastonline = None

        # Check for device subtype
        if device.id_device_subtype == 0:
            device.id_device_subtype = None

        super().insert(device)
        # Create token
        device.create_token()
        TeraDevice.db().session.commit()

    @classmethod
    def update(cls, update_id: int, values: dict):
        # Check for device subtype
        if 'id_device_subtype' in values:
            if values['id_device_subtype'] == 0:
                values['id_device_subtype'] = None

        # Prevent changes on UUID
        if 'device_uuid' in values:
            del values['device_uuid']

        # Remove object device_subtype
        if 'device_subtype' in values:
            del values['device_subtype']

        super().update(update_id=update_id, values=values)

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        # Safety check - can't delete participants with sessions
        if TeraDeviceParticipant.get_count(filters={'id_device': self.id_device}, with_deleted=with_deleted) > 0:
            return IntegrityError('Device still associated to participant(s)', self.id_device, 't_devices_participants')

        if TeraSessionDevices.get_count(filters={'id_device': self.id_device}, with_deleted=with_deleted) > 0:
            return IntegrityError('Device still has sessions', self.id_device, 't_sessions_devices')

        if TeraSession.get_count(filters={'id_creator_device': self.id_device}, with_deleted=with_deleted) > 0:
            return IntegrityError('Device still has created sessions', self.id_device, 't_sessions')

        if TeraAsset.get_count(filters={'id_device': self.id_device}, with_deleted=with_deleted) > 0:
            return IntegrityError('Device still has created assets', self.id_device, 't_assets')

        if TeraTest.get_count(filters={'id_device': self.id_device}, with_deleted=with_deleted) > 0:
            return IntegrityError('Device still has created tests', self.id_device, 't_tests')

        return None

    def hard_delete_before(self):
        # Delete sessions that we are part of since they will not be deleted otherwise
        for ses in self.device_sessions:
            ses.hard_delete()

    def get_undelete_cascade_relations(self):
        return ['device_service_config']



from opentera.forms.TeraForm import *
from flask_babel import gettext
from opentera.db.models.TeraParticipantGroup import TeraParticipantGroup


class TeraParticipantGroupForm:

    @staticmethod
    def get_participant_group_form(projects: list):
        form = TeraForm("group")

        # Building lists
        project_list = []
        for project in projects:
            project_list.append(TeraFormValue(value_id=project.id_project, value=project.project_name + ' [' +
                                                                                 project.project_site.site_name + ']'))
        # Sections
        section = TeraFormSection("informations", gettext("Information"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_participant_group", gettext("Participant Group ID"), "hidden", True))
        section.add_item(TeraFormItem("participant_group_name", gettext("Participant Group Name"), "text", True))
        section.add_item(TeraFormItem("id_project", gettext("Project"), "array", True, item_values=project_list))
        section.add_item(TeraFormItem("project_name", gettext("Project Name"), "hidden"))

        return form.to_dict()



from opentera.forms.TeraForm import *
from flask_babel import gettext
from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess


class TeraProjectForm:

    @staticmethod
    def get_project_form(accessible_sites: list):
        form = TeraForm("project")

        # Building lists
        sites_list = []
        for site in accessible_sites:
            sites_list.append(TeraFormValue(value_id=site.id_site, value=site.site_name))

        # Sections
        section = TeraFormSection("informations", gettext("Information"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_project", gettext("Project ID"), "hidden", item_required=True))
        section.add_item(TeraFormItem("project_name", gettext("Project Name"), "text", item_required=True))
        section.add_item(TeraFormItem("project_enabled", gettext("Enabled"), "boolean", item_required=True,
                                      item_default=True))
        section.add_item(TeraFormItem("id_site", gettext("Site"), "array", item_required=True, item_values=sites_list))
        section.add_item(TeraFormItem("project_role", gettext("Role"), "hidden"))
        section.add_item(TeraFormItem("site_name", gettext("Site Name"), "hidden"))

        desc_section = TeraFormSection("description", gettext("Description"))
        form.add_section(desc_section)
        desc_section.add_item(TeraFormItem("project_description", gettext("Description"), "longtext",
                                           item_required=False))

        return form.to_dict()



from sqlalchemy.inspection import inspect
from sqlalchemy import text


class SoftInsertMixin:

    @classmethod
    # Returns the id of the object inserted
    def soft_insert(cls, db_object):
        primary_key_name = inspect(db_object.__class__).primary_key[0].name
        # Check if we have an already present deleted association and, if so, restores it
        # Get foreign keys for that object
        foreign_keys = list(inspect(db_object.__class__).tables[0].foreign_keys)

        # Check if object already exists
        query = db_object.query.execution_options(include_deleted=True)
        for key in foreign_keys:
            # query = query.filter(key == getattr(db_object, key.parent.name))
            value = getattr(db_object, key.parent.name)
            if value:
                query = query.filter(text(key.parent.name + '=' + str(value)))
            else:
                query = query.filter(text(key.parent.name + '=NULL'))
        item = query.first()

        if item:
            # Undelete that item only
            if getattr(item, 'deleted_at', None):
                setattr(item, 'deleted_at', None)
            cls.commit()
            return item
        else:
            # Insert item
            setattr(db_object, primary_key_name, None)
            cls.db().session.add(db_object)
            cls.commit()
            return db_object

        # item = TeraDeviceParticipant.query_device_participant_for_participant_device(device_id=db_object.id_device,
        #                                                                              participant_id=
        #                                                                              db_object.id_participant,
        #                                                                              with_deleted=True)
        # if item:
        #     if item.deleted_at:
        #         item.deleted_at = None
        #
        # else:
        #     # No existing item found



# Constants
class RedisVars:
    # Variable used to store the User token encryption key
    RedisVar_UserTokenAPIKey = "UserTokenAPIKey"

    # Variable used to store the dynamic device encryption key
    RedisVar_DeviceTokenAPIKey = "DeviceTokenAPIKey"

    # Variable used to store the static device encryption key (used for initial login)
    RedisVar_DeviceStaticTokenAPIKey = "DeviceStaticTokenAPIKey"

    # Variable used to store the dynamic participant encryption key
    RedisVar_ParticipantTokenAPIKey = "ParticipantTokenAPIKey"

    # Variable used to store the static participant encryption key (used for initial login)
    RedisVar_ParticipantStaticTokenAPIKey = "ParticipantStaticTokenAPIKey"

    # Variable used to store the service token encryption key
    RedisVar_ServiceTokenAPIKey = "ServiceTokenAPIKey"

    # Service prefix to append to get the service information associated with a specific service UUID.
    # For example, a "get" to "service.<uuid>" would get the ServiceInfo object for that service
    RedisVar_ServicePrefixKey = "service."

    # User login attempt counter prefix
    RedisVar_UserLoginAttemptKey = "UserLoginAttempts."

    # Participant login attempt counter prefix
    RedisVar_ParticipantLoginAttemptKey = "ParticipantLoginAttempts."

    @classmethod
    def build_service_rpc_topic(cls, service_key) -> str:
        return cls.RedisVar_ServicePrefixKey + service_key + '.rpc'

    @classmethod
    def build_service_message_topic(cls, service_key) -> str:
        return cls.RedisVar_ServicePrefixKey + service_key + '.messages'



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from sqlalchemy import Column, Integer, String, Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraSite(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_sites'
    id_site = Column(Integer, Sequence('id_site_sequence'), primary_key=True, autoincrement=True)
    site_name = Column(String, nullable=False)

    site_devices = relationship("TeraDevice", secondary="t_devices_sites", back_populates="device_sites")
    site_projects = relationship("TeraProject", cascade="delete", passive_deletes=True,
                                 back_populates='project_site', lazy='selectin')

    site_services_roles = relationship("TeraServiceRole", cascade='delete', passive_deletes=True)

    site_services = relationship("TeraService", secondary="t_services_sites", viewonly=True)

    site_sessions_types = relationship("TeraSessionType", secondary="t_sessions_types_sites", viewonly=True)

    site_tests_types = relationship("TeraTestType", secondary="t_tests_types_sites", viewonly=True)

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['site_projects', 'site_devices', 'site_services_roles', 'site_services',
                              'site_sessions_types', 'site_tests_types'])

        return super().to_json(ignore_fields=ignore_fields)

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_site': self.id_site}

    @staticmethod
    def create_defaults(test=False):
        base_site = TeraSite()
        base_site.site_name = 'Default Site'
        TeraSite.insert(base_site)

        if test:
            base_site = TeraSite()
            base_site.site_name = 'Top Secret Site'
            TeraSite.insert(base_site)

    @staticmethod
    def get_site_by_sitename(sitename, with_deleted: bool = False):
        return TeraSite.query.execution_options(include_deleted=with_deleted).filter_by(site_name=sitename).first()

    @staticmethod
    def get_site_by_id(site_id: int, with_deleted: bool = False):
        return TeraSite.query.execution_options(include_deleted=with_deleted).filter_by(id_site=site_id).first()

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        for project in self.site_projects:
            cannot_be_deleted_exception = project.delete_check_integrity(with_deleted=with_deleted)
            if cannot_be_deleted_exception:
                return IntegrityError('Still have projects with participants with sessions', self.id_site,
                                      't_projects')
        return None

    @classmethod
    def insert(cls, site):
        # Creates admin and user roles for that site
        super().insert(site)

        from opentera.db.models.TeraServiceRole import TeraServiceRole
        from opentera.db.models.TeraService import TeraService
        opentera_service_id = TeraService.get_openteraserver_service().id_service
        access_role = TeraServiceRole()
        access_role.id_service = opentera_service_id
        access_role.id_site = site.id_site
        access_role.service_role_name = 'admin'
        TeraServiceRole.insert(access_role)

        access_role = TeraServiceRole()
        access_role.id_service = opentera_service_id
        access_role.id_site = site.id_site
        access_role.service_role_name = 'user'
        TeraServiceRole.insert(access_role)

    def get_undelete_cascade_relations(self) -> list:
        return ['site_services_roles']



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship
from datetime import datetime
import json
import jsonschema


class TeraServiceConfigSpecific(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_services_configs_specifics'
    id_service_config_specific = Column(Integer, Sequence('id_service_config_specific_sequence'),
                                        primary_key=True, autoincrement=True)
    id_service_config = Column(Integer, ForeignKey('t_services_configs.id_service_config', ondelete='cascade'),
                               nullable=False)
    service_config_specific_id = Column(String, nullable=False)
    service_config_specific_config = Column(String, nullable=False)

    service_config_specific_service_config = relationship("TeraServiceConfig", viewonly=True)

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_config_specific_service_config'])

        if minimal:
            ignore_fields.extend([])

        json_config = super().to_json(ignore_fields=ignore_fields)

        return json_config

    @staticmethod
    def get_service_config_specific_by_id(s_id: int, with_deleted: bool = False):
        return TeraServiceConfigSpecific.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service_config_specific=s_id).first()

    @staticmethod
    def get_service_config_specifics_for_service_config(service_config_id: int, with_deleted: bool = False):
        return TeraServiceConfigSpecific.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service_config=service_config_id).all()

    def get_last_update_datetime(self) -> datetime:
        return datetime.fromtimestamp(self.version_id/1000)

    @staticmethod
    def insert_or_update_specific_config(service_config_id: int, specific_id: str, config: str):
        # Check if we have an existing specific config for that service_config
        configs = TeraServiceConfigSpecific.get_service_config_specifics_for_service_config(service_config_id)
        insertion = True

        if configs:
            for s_config in configs:
                if s_config.service_config_specific_id == specific_id:
                    # Update
                    s_config.service_config_specific_config = config
                    insertion = False
                    break
        if insertion:
            # Insert
            new_specific_config = TeraServiceConfigSpecific()
            new_specific_config.id_service_config = service_config_id
            new_specific_config.service_config_specific_id = specific_id
            new_specific_config.service_config_specific_config = config
            TeraServiceConfigSpecific.db().session.add(new_specific_config)
        TeraServiceConfigSpecific.db().session.commit()



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
import uuid
from sqlalchemy import or_
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, TIMESTAMP
from sqlalchemy.orm import relationship


class TeraAsset(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_assets'
    id_asset = Column(Integer, Sequence('id_asset_sequence'), primary_key=True, autoincrement=True)
    id_session = Column(Integer, ForeignKey("t_sessions.id_session", ondelete='cascade'), nullable=False)
    # Creator of that asset - multiple could be used to indicate, for example, an asset created by a device for a
    # specific participant in a session
    id_device = Column(Integer, ForeignKey("t_devices.id_device"), nullable=True)
    id_participant = Column(Integer, ForeignKey("t_participants.id_participant"), nullable=True)
    id_user = Column(Integer, ForeignKey("t_users.id_user"), nullable=True)
    id_service = Column(Integer, ForeignKey("t_services.id_service"), nullable=True)
    # Put a description of the asset here
    asset_name = Column(String, nullable=False)

    asset_uuid = Column(String(36), nullable=False, unique=True)
    asset_service_uuid = Column(String(36), ForeignKey("t_services.service_uuid", ondelete='cascade'), nullable=False)
    asset_type = Column(String, nullable=False)  # MIME Type
    asset_datetime = Column(TIMESTAMP(timezone=True), nullable=True)

    asset_session = relationship("TeraSession", back_populates='session_assets')
    asset_device = relationship("TeraDevice", back_populates='device_assets')
    asset_user = relationship("TeraUser", back_populates='user_assets')
    asset_participant = relationship("TeraParticipant", back_populates='participant_assets')
    asset_service = relationship("TeraService", foreign_keys="TeraAsset.id_service")
    asset_service_owner = relationship("TeraService", foreign_keys="TeraAsset.asset_service_uuid")

    def from_json(self, json, ignore_fields=None):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['asset_session', 'asset_device', 'asset_user', 'asset_participant', 'asset_service',
                              'asset_service_owner', 'asset_service_owner_name', 'asset_session_name'])

        super().from_json(json, ignore_fields)

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['asset_session', 'asset_device', 'asset_user', 'asset_participant', 'asset_service',
                              'asset_service_owner', 'asset_service_owner_name', 'asset_session_name'])

        asset_json = super().to_json(ignore_fields=ignore_fields)
        if not minimal:
            asset_json['asset_service_owner_name'] = self.asset_service_owner.service_name
            asset_json['asset_session_name'] = self.asset_session.session_name
            if self.id_device:
                asset_json['asset_device'] = self.asset_device.device_name
            if self.id_user:
                asset_json['asset_user'] = self.asset_user.get_fullname()
            if self.id_participant:
                asset_json['asset_participant'] = self.asset_participant.participant_name
            if self.id_service:
                asset_json['asset_service'] = self.asset_service.service_name

        return asset_json

    def to_json_create_event(self):
        return self.to_json(minimal=True)

    def to_json_update_event(self):
        return self.to_json(minimal=True)

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_asset': self.id_asset, 'asset_uuid': self.asset_uuid}

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraSession import TeraSession
            from opentera.db.models.TeraDevice import TeraDevice
            from opentera.db.models.TeraParticipant import TeraParticipant
            from opentera.db.models.TeraUser import TeraUser

            session2 = TeraSession.get_session_by_name("Séance #2")
            session3 = TeraSession.get_session_by_name("Séance #3")

            for i in range(3):
                new_asset = TeraAsset()
                new_asset.asset_name = "Asset #" + str(i)
                new_asset.asset_session = session2
                new_asset.asset_uuid = str(uuid.uuid4())
                new_asset.asset_service_uuid = '00000000-0000-0000-0000-000000000001'
                new_asset.asset_type = 'application/octet-stream'
                if i == 0:
                    new_asset.id_participant = TeraParticipant.get_participant_by_name('Participant #1').id_participant
                if i == 1:
                    new_asset.id_user = TeraUser.get_user_by_id(1).id_user
                TeraAsset.insert(new_asset)

            asset_device = TeraDevice.get_device_by_name('Apple Watch #W05P1')
            new_asset = TeraAsset()
            new_asset.asset_name = "Device Asset"
            new_asset.asset_session = session3
            new_asset.asset_device = asset_device
            new_asset.asset_uuid = str(uuid.uuid4())
            new_asset.asset_service_uuid = '00000000-0000-0000-0000-000000000001'
            new_asset.asset_type = 'video/mpeg'
            new_asset.id_service = 1
            TeraAsset.insert(new_asset)

    @staticmethod
    def get_asset_by_id(asset_id: int, with_deleted: bool = False):
        return TeraAsset.query.filter_by(id_asset=asset_id).execution_options(include_deleted=with_deleted).first()

    @staticmethod
    def get_asset_by_uuid(asset_uuid: str, with_deleted: bool = False):
        return TeraAsset.query.filter_by(asset_uuid=asset_uuid).execution_options(include_deleted=with_deleted).first()

    @staticmethod
    def get_assets_for_device(device_id: int, with_deleted: bool = False):
        # This returns all assets that the device has access to
        from opentera.db.models.TeraSession import TeraSession
        return TeraAsset.query.join(TeraSession).execution_options(include_deleted=with_deleted)\
            .filter(or_(TeraSession.session_devices.any(id_device=device_id), TeraAsset.id_device == device_id)).all()

    @staticmethod
    def get_assets_for_user(user_id: int, with_deleted: bool = False):
        # This return all assets that the user has access to
        from opentera.db.models.TeraSession import TeraSession
        return TeraAsset.query.join(TeraSession).execution_options(include_deleted=with_deleted)\
            .filter(or_(TeraSession.session_users.any(id_user=user_id), TeraAsset.id_user == user_id)).all()

    @staticmethod
    def get_assets_for_session(session_id: int, with_deleted: bool = False):
        return TeraAsset.query.execution_options(include_deleted=with_deleted).filter_by(id_session=session_id).all()

    @staticmethod
    def get_assets_for_participant(part_id: int, with_deleted: bool = False):
        # This returns all assets that the participant has access
        from opentera.db.models.TeraSession import TeraSession
        return TeraAsset.query.join(TeraSession).execution_options(include_deleted=with_deleted).\
            filter(or_(TeraSession.session_participants.any(id_participant=part_id),
                       TeraAsset.id_participant == part_id)).all()

    @staticmethod
    def get_assets_owned_by_service(service_uuid: str, with_deleted: bool = False):
        return TeraAsset.query.execution_options(include_deleted=with_deleted).\
            filter_by(asset_service_uuid=service_uuid).all()

    @staticmethod
    def get_assets_created_by_service(service_id: int, with_deleted: bool = False):
        return TeraAsset.query.execution_options(include_deleted=with_deleted).filter_by(id_service=service_id).all()

    @staticmethod
    def get_assets_created_by_user(user_id: int, with_deleted: bool = False):
        return TeraAsset.query.execution_options(include_deleted=with_deleted).filter_by(id_user=user_id).all()

    @staticmethod
    def get_assets_created_by_participant(participant_id: int, with_deleted: bool = False):
        return TeraAsset.query.execution_options(include_deleted=with_deleted).\
            filter_by(id_participant=participant_id).all()

    @staticmethod
    def get_assets_created_by_device(device_id: int, with_deleted: bool = False):
        return TeraAsset.query.execution_options(include_deleted=with_deleted).filter_by(id_device=device_id).all()

    @staticmethod
    def get_access_token(asset_uuids: list, token_key: str, requester_uuid: str, expiration=3600):
        import time
        import jwt

        # Creating token with user info
        now = time.time()
        payload = {
            'iat': int(now),
            'exp': int(now) + expiration,
            'iss': 'TeraServer',
            'asset_uuids': asset_uuids,
            'requester_uuid': requester_uuid
        }

        return jwt.encode(payload, token_key, algorithm='HS256')

    @classmethod
    def insert(cls, asset):
        # Generate UUID
        asset.asset_uuid = str(uuid.uuid4())

        super().insert(asset)



from opentera.forms.TeraForm import *

from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess
from flask_babel import gettext


class TeraServiceConfigForm:

    @staticmethod
    def get_service_config_form():
        form = TeraForm("service_config")

        # Building lists
        #################
        # None to build!

        # Sections
        section = TeraFormSection("infos", gettext("Information"))
        form.add_section(section)

        # service_endpoint = db.Column(db.String, nullable=False)
        # service_clientendpoint = db.Column(db.String, nullable=False)
        # service_enabled = db.Column(db.Boolean, nullable=False, default=False)
        # Items
        section.add_item(TeraFormItem("id_service", gettext("Service ID"), "hidden", item_required=True))
        section.add_item(TeraFormItem("id_service_config", gettext("Service Config ID"), "hidden", item_required=True))
        section.add_item(TeraFormItem("id_user", gettext("User ID"), "hidden", item_required=False))
        section.add_item(TeraFormItem("id_device", gettext("Device ID"), "hidden", item_required=False))
        section.add_item(TeraFormItem("id_participant", gettext("Participant ID"), "hidden", item_required=False))
        section.add_item(TeraFormItem("service_config_config", gettext("Service Config"), "hidden", item_required=True))

        form_dict = form.to_dict()

        # if service.has_config_schema():
        #     import json
        #     config_json = json.loads(service.service_config_schema)
        #     form_dict.update(config_json)

        return form_dict

    @staticmethod
    def get_service_config_config_form(service_key: str):
        form = TeraForm("service_config_config")

        if service_key == 'VideoRehabService':
            # Sections
            section1 = TeraFormSection("main_audio_video", gettext("Multimedia Configuration"))
            form.add_section(section1)

            # Items
            section1.add_item(TeraFormItem("camera", gettext("Camera"), "videoinputs", False))
            section1.add_item(TeraFormItem("mirror", gettext("Mirrored image"), "boolean", False, item_default=True))

            item = TeraFormItem("teracam_src", gettext("URL"), "text", True,
                                item_condition=TeraFormItemCondition("camera", "=", "OpenTeraCam"))
            section1.add_item(item)

            section1.add_item(TeraFormItem("camera_ptz", gettext("Pan-Tilt-Zoom Camera"), "boolean"))
            item = TeraFormItem("camera_ptz_type", gettext("Control Type"), "array", True,
                                [TeraFormValue("0", gettext("Vivotek")),
                                 # TeraFormValue("1", gettext("ONVIF (Generic)"))
                                 ],
                                item_condition=TeraFormItemCondition("camera_ptz", "=", True))
            section1.add_item(item)
            item = TeraFormItem("camera_ptz_ip", gettext("Network Address"), "text", True,
                                item_condition=TeraFormItemCondition("camera_ptz", "=", True))
            section1.add_item(item)
            item = TeraFormItem("camera_ptz_port", gettext("Port"), "numeric", True,
                                item_condition=TeraFormItemCondition("camera_ptz", "=", True))
            section1.add_item(item)
            item = TeraFormItem("camera_ptz_username", gettext("Username"), "text", True,
                                item_condition=TeraFormItemCondition("camera_ptz", "=", True))
            section1.add_item(item)
            item = TeraFormItem("camera_ptz_password", gettext("Password"), "password", True,
                                item_condition=TeraFormItemCondition("camera_ptz", "=", True))
            section1.add_item(item)

            section1.add_item(TeraFormItem("audio", gettext("Microphone"), "audioinputs", False))
            section1.add_item(TeraFormItem("camera2", gettext("Secondary Camera"), "videoinputs"))

        return form.to_dict()



from opentera.utils.assets.BaseAsset import BaseAsset
from dataclasses import dataclass


@dataclass
class AssetFile(BaseAsset):
    file_size: int = 0
    file_name: str = ''

    @property
    def asset_type(self) -> str:
        # Automatically detect mime type from filename
        if self.file_name:
            import mimetypes
            mime = mimetypes.guess_type(self.file_name)[0]
            if mime:
                return mime
        return 'application/octet-stream'



from opentera.forms.TeraForm import *

from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess
from flask_babel import gettext


class TeraDeviceForm:

    @staticmethod
    def get_device_form():
        form = TeraForm("device")

        # Building lists
        #################
        # Device types & subtypes
        from opentera.db.models.TeraDeviceType import TeraDeviceType
        device_types = TeraDeviceType.get_devices_types()
        device_types_list = []
        for dev_type in device_types:
            name = gettext(dev_type.get_name())
            device_types_list.append(TeraFormValue(value_id=dev_type.id_device_type, value=name))

        # Sections
        section = TeraFormSection("details", gettext("Parameters"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_device", gettext("Device ID"), "hidden", item_required=True))
        section.add_item(TeraFormItem("device_uuid", gettext("Device UUID"), "hidden", item_required=True))
        section.add_item(TeraFormItem("device_name", gettext("Device Name"), "text", item_required=True))
        section.add_item(TeraFormItem("id_device_type", gettext("Device Type ID"), "array", item_required=True,
                                      item_values=device_types_list))
        section.add_item(TeraFormItem("id_device_subtype", gettext("Device Sub-Type"), "array", item_required=False,
                                      item_condition=TeraFormItemCondition("id_device_type", "=", "changed",
                                                                           "/api/user/devicesubtypes?id_device_type=")))
        section.add_item(TeraFormItem("device_token", gettext("Access Token"), "longlabel",
                                      item_options={"readonly": False}))
        section.add_item(TeraFormItem("device_certificate", gettext("Certificate"), "hidden",
                                      item_options={"readonly": True}))
        section.add_item(TeraFormItem("device_enabled", gettext("Device Enabled?"), "boolean", item_required=True))
        section.add_item(TeraFormItem("device_onlineable", gettext("Device Onlineable?"), "boolean",
                                      item_required=True))

        section.add_item(TeraFormItem("device_lastonline", gettext("Last Connection"), "datetime",
                                      item_options={"readonly": True}))

        section3 = TeraFormSection("infos", gettext("Information"))
        form.add_section(section3)
        section3.add_item(TeraFormItem("device_infos", gettext("Device Information"), "longtext"))

        section2 = TeraFormSection("config", gettext("Configuration"))
        form.add_section(section2)
        section2.add_item(TeraFormItem("device_config", gettext("Device Configuration"), "longtext"))

        section1 = TeraFormSection("notes", gettext("Notes"))
        form.add_section(section1)
        section1.add_item(TeraFormItem("device_notes", gettext("Notes"), "longtext"))

        return form.to_dict()
